from __future__ import annotations

import dataclasses
import functools
import re
import sys
import warnings

from contextlib import suppress
from pathlib import Path
from typing import TYPE_CHECKING
from urllib.parse import unquote
from urllib.parse import urlsplit
from urllib.request import url2pathname

from poetry.core.constraints.version import Version
from poetry.core.constraints.version import VersionRange
from poetry.core.constraints.version import parse_marker_version_constraint
from poetry.core.version.markers import SingleMarker
from poetry.core.version.markers import SingleMarkerLike
from poetry.core.version.markers import dnf


if TYPE_CHECKING:
    from poetry.core.constraints.generic import BaseConstraint
    from poetry.core.constraints.version import VersionConstraint
    from poetry.core.version.markers import BaseMarker

    # Even though we've `from __future__ import annotations`, mypy doesn't seem to like
    # this as `dict[str, ...]`
    ConvertedMarkers = dict[str, list[list[tuple[str, str]]]]


BZ2_EXTENSIONS = (".tar.bz2", ".tbz")
XZ_EXTENSIONS = (".tar.xz", ".txz", ".tlz", ".tar.lz", ".tar.lzma")
ZIP_EXTENSIONS = (".zip", ".whl")
TAR_EXTENSIONS = (".tar.gz", ".tgz", ".tar")
ARCHIVE_EXTENSIONS = ZIP_EXTENSIONS + BZ2_EXTENSIONS + TAR_EXTENSIONS + XZ_EXTENSIONS
SUPPORTED_EXTENSIONS: tuple[str, ...] = ZIP_EXTENSIONS + TAR_EXTENSIONS

with suppress(ImportError):
    import bz2  # noqa: F401

    SUPPORTED_EXTENSIONS += BZ2_EXTENSIONS

with suppress(ImportError):
    # Only for Python 3.3+
    import lzma  # noqa: F401

    SUPPORTED_EXTENSIONS += XZ_EXTENSIONS


def path_to_url(path: str | Path) -> str:
    """
    Convert a path to a file: URL.  The path will be made absolute unless otherwise
    specified and have quoted path parts.
    """
    return Path(path).absolute().as_uri()


def url_to_path(url: str) -> Path:
    """
    Convert an RFC8089 file URI to path.

    The logic used here is borrowed from pip
    https://github.com/pypa/pip/blob/4d1932fcdd1974c820ea60b3286984ebb0c3beaa/src/pip/_internal/utils/urls.py#L31
    """
    if not url.startswith("file:"):
        raise ValueError(f"{url} is not a valid file URI")

    _, netloc, path, _, _ = urlsplit(url)

    if not netloc or netloc == "localhost":
        # According to RFC 8089, same as empty authority.
        netloc = ""
    elif netloc not in {".", ".."} and sys.platform == "win32":
        if re.match(r"^(localhost)?([a-zA-Z]:)$", netloc):
            path = f"/{netloc[-2:]}{path}"
            if netloc.startswith("localhost"):
                message = (
                    f"The file URL {url} is missing a slash between localhost"
                    " and the drive letter."
                    f" Did you mean file://localhost{path}?"
                )
            else:
                message = (
                    f"The file URL {url} uses a drive letter without a leading slash."
                    f" Did you mean file://{path}?"
                )
            warnings.warn(message, UserWarning, stacklevel=2)
            netloc = ""
        else:
            # If we have a UNC path, prepend UNC share notation.
            netloc = "\\\\" + netloc
    else:
        raise ValueError(
            f"non-local file URIs are not supported on this platform: {url}"
        )

    return Path(url2pathname(netloc + unquote(path)))


def is_url(name: str) -> bool:
    if ":" not in name:
        return False
    scheme = name.split(":", 1)[0].lower()

    return scheme in [
        "http",
        "https",
        "file",
        "ftp",
        "ssh",
        "git",
        "hg",
        "bzr",
        "sftp",
        "svn",
        "ssh",
    ]


def strip_extras(path: str) -> tuple[Path, str | None]:
    m = re.match(r"^(.+)(\[[^\]]+\])$", path)
    extras = None
    if m:
        path_no_extras = m.group(1)
        extras = m.group(2)
    else:
        path_no_extras = path

    return Path(path_no_extras), extras


@functools.cache
def cached_is_dir(path: Path) -> bool:
    """A cached version of `Path.is_dir`."""
    return path.is_dir()


@functools.cache
def is_python_project(path: Path) -> bool:
    """Return true if the directory is a Python project"""
    if not cached_is_dir(path):
        return False

    setup_py = path / "setup.py"
    setup_cfg = path / "setup.cfg"
    setuptools_project = setup_py.exists() or setup_cfg.exists()

    pyproject = (path / "pyproject.toml").exists()

    return pyproject or setuptools_project


def is_archive_file(name: str | Path) -> bool:
    """Return True if `name` is a considered as an archive file."""
    ext = splitext(name)[1].lower()
    return ext in ARCHIVE_EXTENSIONS


def splitext(path: str | Path) -> tuple[str, str]:
    """Like pathlib.Path.stem and suffix, but take off .tar too"""
    if isinstance(path, str):
        path = Path(path)
    base, ext = path.stem, path.suffix
    if base.lower().endswith(".tar"):
        ext = f"{base[-4:]}{ext}"
        base = base[:-4]
    return base, ext


def convert_markers(marker: BaseMarker) -> ConvertedMarkers:
    from poetry.core.version.markers import MarkerUnion
    from poetry.core.version.markers import MultiMarker
    from poetry.core.version.markers import SingleMarker

    requirements: ConvertedMarkers = {}
    marker = dnf(marker)
    conjunctions = marker.markers if isinstance(marker, MarkerUnion) else [marker]
    group_count = len(conjunctions)

    def add_constraint(
        marker_name: str, constraint: tuple[str, str], group_index: int
    ) -> None:
        # python_full_version is equivalent to python_version
        # for Poetry so we merge them
        if marker_name == "python_full_version":
            marker_name = "python_version"
        if marker_name not in requirements:
            requirements[marker_name] = [[] for _ in range(group_count)]
        requirements[marker_name][group_index].append(constraint)

    for i, sub_marker in enumerate(conjunctions):
        if isinstance(sub_marker, MultiMarker):
            for m in sub_marker.markers:
                assert isinstance(m, SingleMarkerLike)
                if isinstance(m, SingleMarker):
                    add_constraint(m.name, (m.operator, m.value), i)
                else:
                    add_constraint(m.name, ("", str(m.constraint)), i)
        elif isinstance(sub_marker, SingleMarkerLike):
            if isinstance(sub_marker, SingleMarker):
                add_constraint(
                    sub_marker.name, (sub_marker.operator, sub_marker.value), i
                )
            else:
                add_constraint(sub_marker.name, ("", str(sub_marker.constraint)), i)

    for group_name in requirements:
        # remove duplicates
        seen = []
        for r in requirements[group_name]:
            if r not in seen:
                seen.append(r)
        requirements[group_name] = seen

    return requirements


def contains_group_without_marker(markers: ConvertedMarkers, marker_name: str) -> bool:
    return marker_name not in markers or [] in markers[marker_name]


def create_nested_marker(
    name: str,
    constraint: BaseConstraint | VersionConstraint,
) -> str:
    from poetry.core.constraints.generic import Constraint
    from poetry.core.constraints.generic import MultiConstraint
    from poetry.core.constraints.generic import UnionConstraint
    from poetry.core.constraints.version import VersionUnion

    if constraint.is_any():
        return ""

    if isinstance(constraint, (MultiConstraint, UnionConstraint)):
        multi_parts = []
        for c in constraint.constraints:
            multi = isinstance(c, (MultiConstraint, UnionConstraint))
            multi_parts.append((multi, create_nested_marker(name, c)))

        glue = " and "
        if isinstance(constraint, UnionConstraint):
            parts = [f"({part[1]})" if part[0] else part[1] for part in multi_parts]
            glue = " or "
        else:
            parts = [part[1] for part in multi_parts]

        marker = glue.join(parts)
    elif isinstance(constraint, Constraint):
        marker = f'{name} {constraint.operator} "{constraint.value}"'
    elif isinstance(constraint, VersionUnion):
        parts = [create_nested_marker(name, c) for c in constraint.ranges]
        glue = " or "
        parts = [f"({part})" for part in parts]
        marker = glue.join(parts)
    elif isinstance(constraint, Version):
        if name == "python_version" and constraint.precision >= 3:
            name = "python_full_version"

        marker = f'{name} == "{constraint.text}"'
    else:
        assert isinstance(constraint, VersionRange), (
            f"Unexpected constraint of type {type(constraint)}"
        )
        min_name = max_name = name

        parts = []

        # `python_version` is a special case: to keep the constructed marker equivalent
        # to the constraint we need to be careful with the precision.
        #
        # PEP 440 tells us that when we come to make the comparison the release
        # segment will be zero padded: eg "<= 3.10" is equivalent to "<= 3.10.0".
        #
        # But "python_version <= 3.10" is _not_ equivalent to "python_version <= 3.10.0"
        # - see normalize_python_version_markers.
        #
        # A similar issue arises for a constraint like "> 3.6".
        if constraint.min is not None:
            op = ">=" if constraint.include_min else ">"
            version = constraint.min
            if min_name == "python_version" and version.precision >= 3:
                min_name = "python_full_version"

            if (
                min_name == "python_version"
                and not constraint.include_min
                and version.precision < 3
            ):
                release = version.release
                release = dataclasses.replace(
                    release,
                    major=release.major or 0,
                    minor=release.minor or 0,
                    patch=release.patch or 0,
                )
                version = dataclasses.replace(version, release=release)
                part = f'python_full_version > "{version}"'
            else:
                part = f'{min_name} {op} "{version.stable}"'

            parts.append(part)

        if constraint.max is not None:
            op = "<=" if constraint.include_max else "<"
            version = constraint.max
            if max_name == "python_version" and version.precision >= 3:
                max_name = "python_full_version"

            if (
                max_name == "python_version"
                and constraint.include_max
                and version.precision < 3
            ):
                release = version.release
                release = dataclasses.replace(
                    release,
                    major=release.major or 0,
                    minor=release.minor or 0,
                    patch=release.patch or 0,
                )
                version = dataclasses.replace(version, release=release)
                part = f'python_full_version <= "{version}"'
            else:
                part = f'{max_name} {op} "{version.stable}"'

            parts.append(part)

        marker = " and ".join(parts)

    return marker


def get_python_constraint_from_marker(
    marker: BaseMarker,
) -> VersionConstraint:
    from poetry.core.constraints.version import EmptyConstraint
    from poetry.core.constraints.version import VersionRange

    python_marker = marker.only("python_version", "python_full_version")
    if python_marker.is_any():
        return VersionRange()

    if python_marker.is_empty():
        return EmptyConstraint()

    markers = convert_markers(marker)
    if contains_group_without_marker(markers, "python_version"):
        # groups are in disjunctive normal form (DNF),
        # an empty group means that python_version does not appear in this group,
        # which means that python_version is arbitrary for this group
        return VersionRange()

    python_version_markers = markers["python_version"]
    normalized = normalize_python_version_markers(python_version_markers)
    constraint = parse_marker_version_constraint(normalized)
    return constraint


def normalize_python_version_markers(  # NOSONAR
    disjunction: list[list[tuple[str, str]]],
) -> str:
    ors = []
    for or_ in disjunction:
        ands = []
        for op, version in or_:
            # Expand python version
            if op == "==" and "*" not in version and version.count(".") < 2:
                if version.count(".") == 0:
                    version += ".0"
                version = "~" + version
                op = ""

            elif op == "!=" and "*" not in version and version.count(".") < 2:
                if version.count(".") == 0:
                    version += ".0"
                version += ".*"

            elif op in ("<=", ">"):
                # Make adjustments on encountering versions with less than full
                # precision.
                #
                # Per PEP-508:
                # python_version <-> '.'.join(platform.python_version_tuple()[:2])
                #
                # So for two digits of precision we make the following adjustments:
                # - `python_version > "x.y"` requires version >= x.(y+1).anything
                # - `python_version <= "x.y"` requires version < x.(y+1).anything
                #
                # Treatment when we see a single digit of precision is less clear: is
                # that even a legitimate marker?
                #
                # Experiment suggests that pip behaviour is essentially to make a
                # lexicographical comparison, for example `python_version > "3"` is
                # satisfied by version 3.anything, whereas `python_version <= "3"` is
                # satisfied only by version 2.anything.
                #
                # We achieve the above by fiddling with the operator and version in the
                # marker.
                parsed_version = Version.parse(version)
                if parsed_version.precision < 3:
                    if op == "<=":
                        op = "<"
                    elif op == ">":
                        op = ">="

                    version = parsed_version.next_minor().text

            elif op in ("in", "not in"):
                versions = []
                for v in SingleMarker.VALUE_SEPARATOR_RE.split(version):
                    split = v.split(".")
                    if len(split) in (1, 2):
                        split.append("*")
                        op_ = "" if op == "in" else "!="
                    else:
                        op_ = "==" if op == "in" else "!="

                    versions.append(op_ + ".".join(split))

                if versions:
                    glue = " || " if op == "in" else ", "
                    ands.append(glue.join(versions))

                continue

            ands.append(f"{op}{version}")

        ors.append(" ".join(ands))

    return " || ".join(ors)



from __future__ import annotations

import pytest

from poetry.core.version.pep440 import Release
from poetry.core.version.pep440 import ReleaseTag
from poetry.core.version.pep440.segments import RELEASE_PHASE_NORMALIZATIONS


def test_release_post_init_minor_and_patch() -> None:
    """
    Minor and patch must not be None but zero if there are extra parts.
    """
    release = Release(1, extra=(0,))
    assert release.minor == 0
    assert release.patch == 0


def test_release_post_init_zero_version() -> None:
    """
    Smoke test for edge case (because zeros are stripped for comparison).
    """
    Release(0)


@pytest.mark.parametrize("precision1", range(1, 6))
@pytest.mark.parametrize("precision2", range(1, 6))
def test_release_equal_zero_padding(precision1: int, precision2: int) -> None:
    release1 = Release.from_parts(*range(1, precision1 + 1))
    if precision1 > precision2:
        # e.g. 1.2.3 != 1.2
        release2 = Release.from_parts(*range(1, precision2 + 1))
        assert release1 != release2
        assert release2 != release1
    else:
        # e.g. 1.2 == 1.2.0
        release2 = Release.from_parts(
            *range(1, precision1 + 1), *[0] * (precision2 - precision1)
        )
        assert release1 == release2
        assert release2 == release1


@pytest.mark.parametrize(
    "parts,result",
    [
        ((1,), Release(1)),
        ((1, 2), Release(1, 2)),
        ((1, 2, 3), Release(1, 2, 3)),
        ((1, 2, 3, 4), Release(1, 2, 3, (4,))),
        ((1, 2, 3, 4, 5, 6), Release(1, 2, 3, (4, 5, 6))),
    ],
)
def test_release_from_parts_to_parts(parts: tuple[int, ...], result: Release) -> None:
    assert Release.from_parts(*parts) == result
    assert result.to_parts() == parts


@pytest.mark.parametrize("precision", list(range(1, 6)))
def test_release_precision(precision: int) -> None:
    """
    Semantically identical releases might have a different precision, e.g. 1 vs. 1.0
    """
    release = Release.from_parts(1, *[0] * (precision - 1))
    assert release.precision == precision
    assert len(release.to_parts()) == precision


@pytest.mark.parametrize("precision", list(range(1, 6)))
def test_release_text(precision: int) -> None:
    increments = list(range(1, precision + 1))
    zeros = [1] + [0] * (precision - 1)

    assert Release.from_parts(*increments).text == ".".join(str(i) for i in increments)
    assert Release.from_parts(*zeros).text == ".".join(str(i) for i in zeros)


@pytest.mark.parametrize("precision", list(range(1, 6)))
def test_release_next_major(precision: int) -> None:
    release = Release.from_parts(1, *[0] * (precision - 1))
    expected = Release.from_parts(2, *[0] * (precision - 1))
    assert release.next_major() == expected


@pytest.mark.parametrize("precision", list(range(1, 6)))
def test_release_next_minor(precision: int) -> None:
    release = Release.from_parts(1, *[0] * (precision - 1))
    expected = Release.from_parts(1, 1, *[0] * (precision - 2))
    assert release.next_minor() == expected


@pytest.mark.parametrize("precision", list(range(1, 6)))
def test_release_next_patch(precision: int) -> None:
    release = Release.from_parts(1, *[0] * (precision - 1))
    expected = Release.from_parts(1, 0, 1, *[0] * (precision - 3))
    assert release.next_patch() == expected


@pytest.mark.parametrize(
    ("release", "expected"),
    [
        (Release(0), Release(1)),
        (Release(1), Release(2)),
        (Release(0, 0), Release(0, 1)),
        (Release(1, 2), Release(1, 3)),
        (Release(0, 0, 0), Release(0, 0, 1)),
        (Release(1, 2, 3), Release(1, 2, 4)),
        (Release(0, 0, 0, (0,)), Release(0, 0, 0, (1,))),
        (Release(1, 2, 3, (4,)), Release(1, 2, 3, (5,))),
        (Release(0, 0, 0, (0, 0)), Release(0, 0, 0, (0, 1))),
        (Release(1, 2, 3, (4, 5)), Release(1, 2, 3, (4, 6))),
    ],
)
def test_release_next(release: Release, expected: Release) -> None:
    assert release.next() == expected


@pytest.mark.parametrize(
    "parts,result",
    [
        (("a",), ReleaseTag("alpha", 0)),
        (("a", 1), ReleaseTag("alpha", 1)),
        (("b",), ReleaseTag("beta", 0)),
        (("b", 1), ReleaseTag("beta", 1)),
        (("pre",), ReleaseTag("preview", 0)),
        (("pre", 1), ReleaseTag("preview", 1)),
        (("c",), ReleaseTag("rc", 0)),
        (("c", 1), ReleaseTag("rc", 1)),
        (("r",), ReleaseTag("rev", 0)),
        (("r", 1), ReleaseTag("rev", 1)),
    ],
)
def test_release_tag_normalisation(
    parts: tuple[str] | tuple[str, int], result: ReleaseTag
) -> None:
    tag = ReleaseTag(*parts)
    assert tag == result
    assert tag.to_string() == result.to_string()


@pytest.mark.parametrize(
    "parts,result",
    [
        (("a",), ReleaseTag("beta")),
        (("b",), ReleaseTag("rc")),
        (("post",), None),
        (("rc",), None),
        (("rev",), None),
        (("dev",), None),
    ],
)
def test_release_tag_next_phase(parts: tuple[str], result: ReleaseTag | None) -> None:
    assert ReleaseTag(*parts).next_phase() == result


@pytest.mark.parametrize("phase", list({*RELEASE_PHASE_NORMALIZATIONS.keys()}))
def test_release_tag_next(phase: str) -> None:
    tag = ReleaseTag(phase=phase).next()
    assert tag.phase == RELEASE_PHASE_NORMALIZATIONS[phase]
    assert tag.number == 1



from __future__ import annotations

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from pathlib import Path


class Include:
    """
    Represents an "include" entry.

    It can be a glob string, a single file or a directory.

    This class will then detect the type of this include:

        - a package
        - a module
        - a file
        - a directory
    """

    def __init__(self, base: Path, include: str, formats: list[str]) -> None:
        self._base = base
        self._include = str(include)
        self._formats = formats

        self._elements: list[Path] = sorted(self._base.glob(str(self._include)))

    @property
    def base(self) -> Path:
        return self._base

    @property
    def elements(self) -> list[Path]:
        return self._elements

    @property
    def formats(self) -> list[str]:
        return self._formats

    def is_empty(self) -> bool:
        return len(self._elements) == 0

    def refresh(self) -> Include:
        self._elements = sorted(self._base.glob(self._include))

        return self



from __future__ import annotations

import pytest

from poetry.core.constraints.generic import AnyConstraint
from poetry.core.constraints.generic import BaseConstraint
from poetry.core.constraints.generic import Constraint
from poetry.core.constraints.generic import EmptyConstraint
from poetry.core.constraints.generic import MultiConstraint
from poetry.core.constraints.generic import UnionConstraint


def test_allows() -> None:
    c = MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!="))

    assert not c.allows(Constraint("win32"))
    assert not c.allows(Constraint("linux"))
    assert c.allows(Constraint("darwin"))


@pytest.mark.parametrize(
    ("constraint", "expected_any", "expected_all"),
    [
        (EmptyConstraint(), False, True),
        (AnyConstraint(), True, False),
        (Constraint("win32"), False, False),
        (Constraint("linux"), False, False),
        (Constraint("darwin"), True, True),
        (Constraint("win32", "!="), True, False),
        (Constraint("linux", "!="), True, False),
        (Constraint("darwin", "!="), True, False),
        (
            UnionConstraint(Constraint("win32"), Constraint("linux")),
            False,
            False,
        ),
        (
            UnionConstraint(
                Constraint("win32"), Constraint("linux"), Constraint("darwin")
            ),
            True,
            False,
        ),
        (
            UnionConstraint(Constraint("darwin"), Constraint("linux")),
            True,
            False,
        ),
        (
            UnionConstraint(Constraint("darwin"), Constraint("osx")),
            True,
            True,
        ),
        (
            UnionConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
            True,
            False,
        ),
        (
            UnionConstraint(Constraint("darwin", "!="), Constraint("linux", "!=")),
            True,
            False,
        ),
        (
            UnionConstraint(Constraint("darwin", "!="), Constraint("osx", "!=")),
            True,
            False,
        ),
        (
            MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
            True,
            True,
        ),
        (
            MultiConstraint(
                Constraint("win32", "!="),
                Constraint("linux", "!="),
                Constraint("darwin", "!="),
            ),
            True,
            True,
        ),
        (
            MultiConstraint(Constraint("darwin", "!="), Constraint("linux", "!=")),
            True,
            False,
        ),
        (
            MultiConstraint(Constraint("darwin", "!="), Constraint("osx", "!=")),
            True,
            False,
        ),
    ],
)
def test_allows_any_and_allows_all(
    constraint: BaseConstraint, expected_any: bool, expected_all: bool
) -> None:
    c = MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!="))
    assert c.allows_any(constraint) == expected_any
    assert c.allows_all(constraint) == expected_all



from __future__ import annotations

import logging
import sys
import textwrap

from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any


if TYPE_CHECKING:
    from poetry.core.masonry.utils.module import Module
    from poetry.core.poetry import Poetry


METADATA_BASE = """\
Metadata-Version: {metadata_version}
Name: {name}
Version: {version}
Summary: {summary}
"""

logger = logging.getLogger(__name__)


class Builder:
    format: str | None = None

    def __init__(
        self,
        poetry: Poetry,
        executable: Path | None = None,
        config_settings: dict[str, Any] | None = None,
    ) -> None:
        from poetry.core.masonry.metadata import Metadata

        if not poetry.is_package_mode:
            raise RuntimeError(
                "Building a package is not possible in non-package mode."
            )

        self._config_settings = config_settings or {}

        self._poetry = poetry
        self._apply_local_version_label()

        self._package = poetry.package
        self._path: Path = poetry.pyproject_path.parent
        self._excluded_files: set[str] | None = None
        self._executable = Path(executable or sys.executable)
        self._meta = Metadata.from_package(self._package)

    @cached_property
    def _module(self) -> Module:
        from poetry.core.masonry.utils.module import Module

        packages = [
            item
            for item in self._package.packages
            if not self.format or self.format in item["format"]
        ]
        includes = [
            item
            for item in self._package.include
            if not self.format or self.format in item["format"]
        ]

        return Module(
            self._package.name,
            self._path.as_posix(),
            packages=packages,
            includes=includes,
        )

    @property
    def executable(self) -> Path:
        return self._executable

    @property
    def default_target_dir(self) -> Path:
        return self._path / "dist"

    def _apply_local_version_label(self) -> None:
        """Apply local version label from config settings to the poetry package version if present."""
        if local_version_label := self._config_settings.get("local-version"):
            self._poetry.package.version = self._poetry.package.version.replace(
                local=local_version_label
            )

    def build(self, target_dir: Path | None) -> Path:
        raise NotImplementedError

    def find_excluded_files(self, fmt: str | None = None) -> set[str]:
        if self._excluded_files is None:
            from poetry.core.vcs import get_vcs

            # Checking VCS
            vcs = get_vcs(self._path)
            vcs_ignored_files = set(vcs.get_ignored_files()) if vcs else set()

            def add_all_files(path: Path, target: set[str]) -> None:
                all_files = (
                    [f for f in path.glob("**/*") if f.is_file()]
                    if path.is_dir()
                    else [path]
                )
                target.update(f.relative_to(self._path).as_posix() for f in all_files)

            explicitly_excluded: set[str] = set()
            for excluded_glob in self._package.exclude:
                for excluded in self._path.glob(str(excluded_glob)):
                    add_all_files(excluded, explicitly_excluded)

            explicitly_included: set[str] = set()
            for inc in self._module.explicit_includes:
                if fmt and fmt not in inc.formats:
                    continue

                for included in inc.elements:
                    add_all_files(included, explicitly_included)

            ignored = (vcs_ignored_files | explicitly_excluded) - explicitly_included
            for ignored_file in ignored:
                logger.debug(f"Ignoring: {ignored_file}")

            self._excluded_files = ignored

        return self._excluded_files

    def is_excluded(self, filepath: str | Path) -> bool:
        exclude_path = Path(filepath)

        if "__pycache__" in exclude_path.parts or exclude_path.suffix == ".pyc":
            return True

        while True:
            if exclude_path.as_posix() in self.find_excluded_files(fmt=self.format):
                return True

            if len(exclude_path.parts) > 1:
                exclude_path = exclude_path.parent
            else:
                break

        return False

    def find_files_to_add(self, exclude_build: bool = True) -> set[BuildIncludeFile]:
        """
        Finds all files to add to the tarball
        """
        from poetry.core.masonry.utils.package_include import PackageInclude

        to_add = set()

        for include in self._module.includes:
            include.refresh()
            formats = include.formats

            for file in include.elements:
                if "__pycache__" in file.parts:
                    # This is just a shortcut. It will be ignored later anyway.
                    continue

                if (
                    isinstance(include, PackageInclude)
                    and include.source
                    and self.format == "wheel"
                ):
                    source_root = include.base
                else:
                    source_root = self._path

                if (
                    isinstance(include, PackageInclude)
                    and include.target
                    and self.format == "wheel"
                ):
                    target_dir = include.target
                else:
                    target_dir = None

                if file.is_dir():
                    if self.format in formats:
                        for current_file in file.glob("**/*"):
                            include_file = BuildIncludeFile(
                                path=current_file,
                                project_root=self._path,
                                source_root=source_root,
                                target_dir=target_dir,
                            )

                            if not (
                                current_file.is_dir()
                                or self.is_excluded(
                                    include_file.relative_to_project_root()
                                )
                            ):
                                to_add.add(include_file)
                    continue

                include_file = BuildIncludeFile(
                    path=file,
                    project_root=self._path,
                    source_root=source_root,
                    target_dir=target_dir,
                )

                if self.is_excluded(
                    include_file.relative_to_project_root()
                ) and isinstance(include, PackageInclude):
                    continue

                logger.debug(f"Adding: {file}")
                to_add.add(include_file)

        # add build script if it is specified and explicitly required
        if self._package.build_script and not exclude_build:
            to_add.add(
                BuildIncludeFile(
                    path=self._package.build_script,
                    project_root=self._path,
                    source_root=self._path,
                )
            )

        return to_add

    def get_metadata_content(self) -> str:
        content = METADATA_BASE.format(
            metadata_version=self._meta.metadata_version,
            name=self._meta.name,
            version=self._meta.version,
            summary=str(self._meta.summary),
        )

        if self._meta.license:
            assert self._meta.license_expression is None
            license_field = "License: "
            # Indentation is not only for readability, but required
            # so that the line break is not treated as end of field.
            # The exact indentation does not matter,
            # but it is essential to also indent empty lines.
            escaped_license = textwrap.indent(
                self._meta.license, " " * len(license_field), lambda line: True
            ).strip()
            content += f"{license_field}{escaped_license}\n"
        elif self._meta.license_expression:
            content += f"License-Expression: {self._meta.license_expression}\n"

        for license_file in self._meta.license_files:
            content += f"License-File: {license_file}\n"

        if self._meta.keywords:
            content += f"Keywords: {self._meta.keywords}\n"

        if self._meta.author:
            content += f"Author: {self._meta.author}\n"

        if self._meta.author_email:
            content += f"Author-email: {self._meta.author_email}\n"

        if self._meta.maintainer:
            content += f"Maintainer: {self._meta.maintainer}\n"

        if self._meta.maintainer_email:
            content += f"Maintainer-email: {self._meta.maintainer_email}\n"

        if self._meta.requires_python:
            content += f"Requires-Python: {self._meta.requires_python}\n"

        for classifier in self._meta.classifiers:
            content += f"Classifier: {classifier}\n"

        for extra in sorted(self._meta.provides_extra):
            content += f"Provides-Extra: {extra}\n"

        for dep in sorted(self._meta.requires_dist):
            content += f"Requires-Dist: {dep}\n"

        for url in sorted(self._meta.project_urls, key=lambda u: u[0]):
            content += f"Project-URL: {url}\n"

        if self._meta.description_content_type:
            content += (
                f"Description-Content-Type: {self._meta.description_content_type}\n"
            )

        if self._meta.description is not None:
            content += f"\n{self._meta.description}\n"

        return content

    def convert_entry_points(self) -> dict[str, list[str]]:
        result: dict[str, list[str]] = {}

        for group_name, group in self._poetry.package.entry_points.items():
            if group_name == "console-scripts":
                group_name = "console_scripts"
            elif group_name == "gui-scripts":
                group_name = "gui_scripts"
            result[group_name] = sorted(
                f"{name} = {specification}" for name, specification in group.items()
            )

        return result

    def convert_script_files(self) -> list[Path]:
        script_files: list[Path] = []

        for name, specification in self._poetry.local_config.get("scripts", {}).items():
            if isinstance(specification, dict) and specification.get("type") == "file":
                source = specification["reference"]

                if Path(source).is_absolute():
                    raise RuntimeError(
                        f"{source} in {name} is an absolute path. Expected relative"
                        " path."
                    )

                abs_path = Path.joinpath(self._path, source)

                if not self._package.build_script:
                    # scripts can be generated by build_script, in this case they do not exist here
                    if not abs_path.exists():
                        raise RuntimeError(
                            f"{abs_path} in script specification ({name}) is not found."
                        )
                    if not abs_path.is_file():
                        raise RuntimeError(
                            f"{abs_path} in script specification ({name}) is not a file."
                        )

                script_files.append(abs_path)

        return script_files

    def _get_legal_files(self) -> set[Path]:
        return {self._path / f for f in self._meta.license_files}


class BuildIncludeFile:
    def __init__(
        self,
        path: Path | str,
        project_root: Path | str,
        source_root: Path | str,
        target_dir: Path | str | None = None,
    ) -> None:
        """
        :param project_root: the full path of the project's root
        :param path: a full path to the file to be included
        :param source_root: the full root path to resolve to
        :param target_dir: the relative target root to resolve to
        """
        self.path = Path(path)
        self.project_root = Path(project_root).resolve()
        self.source_root = Path(source_root).resolve()
        self.target_dir = None if not target_dir else Path(target_dir)
        if not self.path.is_absolute():
            self.path = self.source_root / self.path
        else:
            self.path = self.path

        self.path = self.path.resolve()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BuildIncludeFile):
            return False

        return self.path == other.path

    def __hash__(self) -> int:
        return hash(self.path)

    def __repr__(self) -> str:
        return str(self.path)

    def relative_to_project_root(self) -> Path:
        return self.path.relative_to(self.project_root)

    def relative_to_source_root(self) -> Path:
        return self.path.relative_to(self.source_root)

    def relative_to_target_root(self) -> Path:
        path = self.relative_to_source_root()
        if self.target_dir is not None:
            return self.target_dir / path
        return path



from __future__ import annotations

import logging

from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

from packaging.utils import canonicalize_name

from poetry.core.packages.dependency import Dependency
from poetry.core.packages.directory_dependency import DirectoryDependency
from poetry.core.packages.file_dependency import FileDependency
from poetry.core.pyproject.toml import PyProjectTOML


logger = logging.getLogger(__name__)


if TYPE_CHECKING:
    from poetry.core.packages.project_package import ProjectPackage


class Poetry:
    def __init__(
        self,
        file: Path,
        local_config: dict[str, Any],
        package: ProjectPackage,
        pyproject_type: type[PyProjectTOML] = PyProjectTOML,
    ) -> None:
        self._pyproject = pyproject_type(file)
        self._package = package
        self._local_config = local_config
        self._build_system_dependencies: list[Dependency] | None = None

    @property
    def pyproject(self) -> PyProjectTOML:
        return self._pyproject

    @property
    def pyproject_path(self) -> Path:
        return self._pyproject.path

    @property
    def package(self) -> ProjectPackage:
        return self._package

    @property
    def is_package_mode(self) -> bool:
        package_mode = self._local_config["package-mode"]
        assert isinstance(package_mode, bool)
        return package_mode

    @property
    def local_config(self) -> dict[str, Any]:
        return self._local_config

    def get_project_config(self, config: str, default: Any = None) -> Any:
        return self._local_config.get("config", {}).get(config, default)

    @property
    def build_system_dependencies(self) -> list[Dependency]:
        if self._build_system_dependencies is None:
            build_system = self.pyproject.build_system
            self._build_system_dependencies = []

            for requirement in build_system.requires:
                dependency = None
                try:
                    dependency = Dependency.create_from_pep_508(requirement)
                except ValueError:
                    # PEP 517 requires can be path if not PEP 508
                    path = Path(requirement)

                    if path.is_file():
                        dependency = FileDependency(
                            name=canonicalize_name(path.name), path=path
                        )
                    elif path.is_dir():
                        dependency = DirectoryDependency(
                            name=canonicalize_name(path.name), path=path
                        )

                # skip since we could not determine requirement
                if dependency:
                    self._build_system_dependencies.append(dependency)
                else:
                    logger.debug(
                        "Skipping build system dependency - could not determine requirement type: %s",
                        requirement,
                    )

        return self._build_system_dependencies



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING

from packaging.utils import NormalizedName
from packaging.utils import canonicalize_name

from poetry.core.version.markers import parse_marker


if TYPE_CHECKING:
    from poetry.core.packages.dependency import Dependency
    from poetry.core.version.markers import BaseMarker

MAIN_GROUP: NormalizedName = canonicalize_name("main")


class DependencyGroup:
    def __init__(
        self, name: str, *, optional: bool = False, mixed_dynamic: bool = False
    ) -> None:
        self._name: NormalizedName = canonicalize_name(name)
        self._pretty_name: str = name
        self._optional: bool = optional
        self._mixed_dynamic = mixed_dynamic
        self._dependencies: list[Dependency] = []
        self._poetry_dependencies: list[Dependency] = []
        self._included_dependency_groups: dict[NormalizedName, DependencyGroup] = {}

    @property
    def name(self) -> NormalizedName:
        return self._name

    @property
    def pretty_name(self) -> str:
        return self._pretty_name

    @property
    def dependencies(self) -> list[Dependency]:
        dependencies = self._dependencies
        included_group_dependencies = self._resolve_included_dependency_groups(
            dependencies_for_locking=False
        )

        if not dependencies:
            # legacy mode
            dependencies = self._poetry_dependencies
        elif self._mixed_dynamic and self._poetry_dependencies:
            if all(dep.is_optional() for dep in self._dependencies):
                dependencies = [
                    *self._dependencies,
                    *(d for d in self._poetry_dependencies if not d.is_optional()),
                ]
            elif all(not dep.is_optional() for dep in self._dependencies):
                dependencies = [
                    *self._dependencies,
                    *(d for d in self._poetry_dependencies if d.is_optional()),
                ]

        return dependencies + included_group_dependencies

    @property
    def dependencies_for_locking(self) -> list[Dependency]:
        included_group_dependencies = self._resolve_included_dependency_groups(
            dependencies_for_locking=True
        )

        if not self._poetry_dependencies:
            dependencies = self._dependencies
        elif not self._dependencies:
            dependencies = self._poetry_dependencies
        else:
            poetry_dependencies_by_name = defaultdict(list)
            for dep in self._poetry_dependencies:
                poetry_dependencies_by_name[dep.name].append(dep)

            dependencies = []
            for dep in self.dependencies:
                if dep.name in poetry_dependencies_by_name:
                    enriched = False
                    dep_marker = dep.marker
                    if dep.in_extras:
                        dep_marker = dep.marker.intersect(
                            parse_marker(
                                " or ".join(
                                    f"extra == '{extra}'" for extra in dep.in_extras
                                )
                            )
                        )
                    for poetry_dep in poetry_dependencies_by_name[dep.name]:
                        marker = dep_marker.intersect(poetry_dep.marker)
                        if not marker.is_empty():
                            if marker == dep_marker:
                                marker = dep.marker
                            enriched = True
                            dependencies.append(
                                _enrich_dependency(dep, poetry_dep, marker)
                            )
                    if not enriched:
                        dependencies.append(dep)
                else:
                    dependencies.append(dep)

        return dependencies + included_group_dependencies

    def _resolve_included_dependency_groups(
        self, *, dependencies_for_locking: bool = False
    ) -> list[Dependency]:
        """Resolves and returns the dependencies from included dependency groups.

        This method iterates over all included dependency groups and collects
        their dependencies, associating them with the current group.
        """
        return [
            dependency.with_groups([self.name])
            for dependency_group in self._included_dependency_groups.values()
            for dependency in (
                dependency_group.dependencies_for_locking
                if dependencies_for_locking
                else dependency_group.dependencies
            )
        ]

    def is_optional(self) -> bool:
        return self._optional

    def add_dependency(self, dependency: Dependency) -> None:
        if not self._dependencies and self._poetry_dependencies:
            self._poetry_dependencies.append(dependency)
        else:
            self._dependencies.append(dependency)

    def add_poetry_dependency(self, dependency: Dependency) -> None:
        self._poetry_dependencies.append(dependency)

    def remove_dependency(self, name: str) -> None:
        from packaging.utils import canonicalize_name

        name = canonicalize_name(name)

        dependencies = []
        for dependency in self.dependencies:
            if dependency.name == name:
                continue
            dependencies.append(dependency)
        self._dependencies = dependencies

        dependencies = []
        for dependency in self._poetry_dependencies:
            if dependency.name == name:
                continue
            dependencies.append(dependency)
        self._poetry_dependencies = dependencies

    def include_dependency_group(self, dependency_group: DependencyGroup) -> None:
        if dependency_group.name == self.name:
            raise ValueError("Cannot include the dependency group to itself.")
        if dependency_group.name in self._included_dependency_groups:
            raise ValueError(
                f"Dependency group {dependency_group.pretty_name} is already included"
            )

        self._included_dependency_groups[dependency_group.name] = dependency_group

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DependencyGroup):
            return NotImplemented

        return (
            self._name == other.name
            and set(self._dependencies) == set(other.dependencies)
            and set(self._poetry_dependencies) == set(other._poetry_dependencies)
        )

    def __repr__(self) -> str:
        cls = self.__class__.__name__
        return (
            f"{cls}({self._pretty_name!r},"
            f" optional={self._optional},"
            f" mixed_dynamic={self._mixed_dynamic})"
        )


def _enrich_dependency(
    project_dependency: Dependency, poetry_dependency: Dependency, marker: BaseMarker
) -> Dependency:
    if (
        project_dependency.source_type is not None
        and poetry_dependency.source_type is not None
        and not poetry_dependency.is_same_source_as(project_dependency)
    ):
        raise ValueError(
            "Cannot enrich dependency with different sources: "
            f"{project_dependency} and {poetry_dependency}"
        )

    constraint = project_dependency.constraint.intersect(poetry_dependency.constraint)
    if constraint.is_empty():
        raise ValueError(
            "Cannot enrich dependency with incompatible constraints: "
            f"{project_dependency} and {poetry_dependency}"
        )

    if project_dependency.source_type is not None:
        from poetry.core.packages.directory_dependency import DirectoryDependency
        from poetry.core.packages.vcs_dependency import VCSDependency

        dependency = project_dependency.clone()
        if isinstance(project_dependency, (DirectoryDependency, VCSDependency)):
            dependency._develop = poetry_dependency._develop
    else:
        dependency = poetry_dependency.with_features(project_dependency.features)
        dependency._optional = project_dependency.is_optional()
        dependency._in_extras = project_dependency.in_extras

    dependency.constraint = constraint
    dependency.marker = marker

    return dependency



from __future__ import annotations

import itertools

from poetry.core.constraints.generic import AnyConstraint
from poetry.core.constraints.generic.base_constraint import BaseConstraint
from poetry.core.constraints.generic.constraint import Constraint
from poetry.core.constraints.generic.constraint import ExtraConstraint
from poetry.core.constraints.generic.empty_constraint import EmptyConstraint
from poetry.core.constraints.generic.multi_constraint import ExtraMultiConstraint
from poetry.core.constraints.generic.multi_constraint import MultiConstraint


class UnionConstraint(BaseConstraint):
    def __init__(self, *constraints: BaseConstraint) -> None:
        self._constraints = constraints

    @property
    def constraints(self) -> tuple[BaseConstraint, ...]:
        return self._constraints

    def allows(
        self,
        other: BaseConstraint,
    ) -> bool:
        return any(constraint.allows(other) for constraint in self._constraints)

    def allows_any(self, other: BaseConstraint) -> bool:
        if isinstance(other, UnionConstraint):
            return any(
                c1.allows_any(c2)
                for c1 in self._constraints
                for c2 in other.constraints
            )

        return any(c.allows_any(other) for c in self._constraints)

    def allows_all(self, other: BaseConstraint) -> bool:
        if isinstance(other, UnionConstraint):
            return all(
                any(c1.allows_all(c2) for c1 in self._constraints)
                for c2 in other.constraints
            )

        return any(c.allows_all(other) for c in self._constraints)

    def invert(self) -> MultiConstraint:
        inverted_constraints = [c.invert() for c in self._constraints]
        if any(not isinstance(c, Constraint) for c in inverted_constraints):
            raise NotImplementedError(
                "Inversion of complex union constraints not implemented"
            )
        if any(isinstance(c, ExtraConstraint) for c in inverted_constraints):
            multi_type: type[MultiConstraint] = ExtraMultiConstraint
        else:
            multi_type = MultiConstraint
        return multi_type(*inverted_constraints)  # type: ignore[arg-type]

    def intersect(self, other: BaseConstraint) -> BaseConstraint:
        if other.is_any():
            return self

        if other.is_empty():
            return other

        if isinstance(other, UnionConstraint) and set(other.constraints) == set(
            self._constraints
        ):
            return self

        if isinstance(other, ExtraConstraint) and other in self._constraints:
            return other

        if isinstance(other, Constraint):
            # (A or B) and C => (A and C) or (B and C)
            # just a special case of UnionConstraint
            other = UnionConstraint(other)

        new_constraints = []
        seen_multi_constraints = set()

        def add_unseen_constraint(constraint: BaseConstraint) -> None:
            if not (
                constraint.is_empty()
                or constraint in new_constraints
                or (
                    isinstance(constraint, MultiConstraint)
                    and frozenset(constraint.constraints) in seen_multi_constraints
                )
            ):
                new_constraints.append(constraint)
                if isinstance(constraint, MultiConstraint):
                    seen_multi_constraints.add(frozenset(constraint.constraints))

        if isinstance(other, UnionConstraint):
            # (A or B) and (A or B or C) => A or B
            our_constraints = set(self._constraints)
            their_constraints = set(other.constraints)
            if our_constraints.issubset(their_constraints):
                return self
            if their_constraints.issubset(our_constraints):
                if len(other.constraints) == 1:
                    return other.constraints[0]
                return other

            # (A or B) and (C or D) => (A and C) or (A and D) or (B and C) or (B and D)
            for our_constraint in self._constraints:
                for their_constraint in other.constraints:
                    add_unseen_constraint(our_constraint.intersect(their_constraint))

        else:
            assert isinstance(other, MultiConstraint)
            # (A or B) and (C and D) => (A and C and D) or (B and C and D)
            # (A or B) and (A and D) => A and D
            for our_constraint in self._constraints:
                intersection = our_constraint
                for their_constraint in other.constraints:
                    intersection = intersection.intersect(their_constraint)
                add_unseen_constraint(intersection)

        if not new_constraints:
            return EmptyConstraint()

        if len(new_constraints) == 1:
            return new_constraints[0]

        return UnionConstraint(*new_constraints)

    def union(self, other: BaseConstraint) -> BaseConstraint:
        if other.is_any():
            return other

        if other.is_empty():
            return self

        if other == self:
            return self

        if isinstance(other, Constraint):
            # (A or B) or C => A or B or C
            # just a special case of UnionConstraint
            other = UnionConstraint(other)

        if isinstance(other, UnionConstraint):
            # (A or B) or (C or D) => A or B or C or D
            our_new_constraints: list[BaseConstraint] = []
            their_new_constraints: list[BaseConstraint] = []
            merged_new_constraints: list[BaseConstraint] = []
            for their_constraint in other.constraints:
                for our_constraint in self._constraints:
                    union = our_constraint.union(their_constraint)
                    if union.is_any():
                        return AnyConstraint()
                    if isinstance(union, Constraint):
                        if union == our_constraint:
                            if union not in our_new_constraints:
                                our_new_constraints.append(union)
                        elif union == their_constraint:
                            if union not in their_new_constraints:
                                their_new_constraints.append(their_constraint)
                        elif union not in merged_new_constraints:
                            merged_new_constraints.append(union)
                    else:
                        if our_constraint not in our_new_constraints:
                            our_new_constraints.append(our_constraint)
                        if their_constraint not in their_new_constraints:
                            their_new_constraints.append(their_constraint)
            new_constraints = our_new_constraints
            for constraint in itertools.chain(
                their_new_constraints, merged_new_constraints
            ):
                if constraint not in new_constraints:
                    new_constraints.append(constraint)

        else:
            assert isinstance(other, MultiConstraint)
            # (A or B) or (A and D) => A or B
            if any(c in other.constraints for c in self._constraints):
                return self

            # (A or B) or (not A and D) => A or B or D
            simplified = False
            our_simple_constraints = [
                c for c in self._constraints if isinstance(c, Constraint)
            ]
            their_new_constraints = []
            for their_constraint in other.constraints:
                if any(
                    c.union(their_constraint).is_any() for c in our_simple_constraints
                ):
                    simplified = True
                else:
                    their_new_constraints.append(their_constraint)
            if simplified:
                if not their_new_constraints:
                    return AnyConstraint()
                return self.union(UnionConstraint(*their_new_constraints))

            # (A or B) or (C and D) => nothing to do
            new_constraints = [*self._constraints, other]

        if len(new_constraints) == 1:
            return new_constraints[0]

        return UnionConstraint(*new_constraints)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, UnionConstraint):
            return False

        return self._constraints == other._constraints

    def __hash__(self) -> int:
        return hash(("union", *self._constraints))

    def __str__(self) -> str:
        constraints = [str(constraint) for constraint in self._constraints]
        return " || ".join(constraints)



from __future__ import annotations

from contextlib import suppress
from typing import TYPE_CHECKING
from typing import Any

from poetry.core.pyproject.tables import BuildSystem
from poetry.core.utils._compat import tomllib


if TYPE_CHECKING:
    from pathlib import Path


class PyProjectTOML:
    def __init__(self, path: Path) -> None:
        self._path = path
        self._data: dict[str, Any] | None = None
        self._build_system: BuildSystem | None = None

    @property
    def path(self) -> Path:
        return self._path

    @property
    def data(self) -> dict[str, Any]:
        if self._data is None:
            if not self.path.exists():
                self._data = {}
            else:
                try:
                    with self.path.open("rb") as f:
                        self._data = tomllib.load(f)
                except tomllib.TOMLDecodeError as e:
                    from poetry.core.pyproject.exceptions import PyProjectError

                    msg = (
                        f"{self._path.as_posix()} is not a valid TOML file.\n"
                        f"{e.__class__.__name__}: {e}"
                    )

                    if str(e).startswith("Cannot overwrite a value"):
                        msg += "\nThis is often caused by a duplicate entry."

                    raise PyProjectError(msg) from e

        return self._data

    @property
    def build_system(self) -> BuildSystem:
        if self._build_system is None:
            build_backend = None
            requires = None

            if not self.path.exists():
                build_backend = "poetry.core.masonry.api"
                requires = ["poetry-core"]

            container = self.data.get("build-system", {})
            self._build_system = BuildSystem(
                build_backend=container.get("build-backend", build_backend),
                requires=container.get("requires", requires),
            )

        return self._build_system

    @property
    def poetry_config(self) -> dict[str, Any]:
        try:
            tool = self.data["tool"]
            assert isinstance(tool, dict)
            config = tool["poetry"]
            assert isinstance(config, dict)
            return config
        except KeyError as e:
            from poetry.core.pyproject.exceptions import PyProjectError

            raise PyProjectError(
                f"[tool.poetry] section not found in {self._path.as_posix()}"
            ) from e

    def is_poetry_project(self) -> bool:
        from poetry.core.pyproject.exceptions import PyProjectError

        if self.path.exists():
            with suppress(PyProjectError):
                _ = self.poetry_config
                return True

            # Even if there is no [tool.poetry] section, a project can still be a
            # valid Poetry project if there is a name and a version in [project]
            # and there are no dynamic fields.
            with suppress(KeyError):
                project = self.data["project"]
                if (
                    project["name"]
                    and project["version"]
                    and not project.get("dynamic")
                ):
                    return True

        return False



"Provides for superficial grammar analysis."

from collections import Counter, defaultdict
from typing import List, Dict, Iterator, FrozenSet, Set

from ..utils import bfs, fzset, classify, OrderedSet
from ..exceptions import GrammarError
from ..grammar import Rule, Terminal, NonTerminal, Symbol
from ..common import ParserConf


class RulePtr:
    __slots__ = ('rule', 'index')
    rule: Rule
    index: int

    def __init__(self, rule: Rule, index: int):
        assert isinstance(rule, Rule)
        assert index <= len(rule.expansion)
        self.rule = rule
        self.index = index

    def __repr__(self):
        before = [x.name for x in self.rule.expansion[:self.index]]
        after = [x.name for x in self.rule.expansion[self.index:]]
        return '<%s : %s * %s>' % (self.rule.origin.name, ' '.join(before), ' '.join(after))

    @property
    def next(self) -> Symbol:
        return self.rule.expansion[self.index]

    def advance(self, sym: Symbol) -> 'RulePtr':
        assert self.next == sym
        return RulePtr(self.rule, self.index+1)

    @property
    def is_satisfied(self) -> bool:
        return self.index == len(self.rule.expansion)

    def __eq__(self, other) -> bool:
        if not isinstance(other, RulePtr):
            return NotImplemented
        return self.rule == other.rule and self.index == other.index

    def __hash__(self) -> int:
        return hash((self.rule, self.index))


State = FrozenSet[RulePtr]

# state generation ensures no duplicate LR0ItemSets
class LR0ItemSet:
    __slots__ = ('kernel', 'closure', 'transitions', 'lookaheads')

    kernel: State
    closure: State
    transitions: Dict[Symbol, 'LR0ItemSet']
    lookaheads: Dict[Symbol, Set[Rule]]

    def __init__(self, kernel, closure):
        self.kernel = fzset(kernel)
        self.closure = fzset(closure)
        self.transitions = {}
        self.lookaheads = defaultdict(set)

    def __repr__(self):
        return '{%s | %s}' % (', '.join([repr(r) for r in self.kernel]), ', '.join([repr(r) for r in self.closure]))


def update_set(set1, set2):
    if not set2 or set1 > set2:
        return False

    copy = set(set1)
    set1 |= set2
    return set1 != copy

def calculate_sets(rules):
    """Calculate FOLLOW sets.

    Adapted from: http://lara.epfl.ch/w/cc09:algorithm_for_first_and_follow_sets"""
    symbols = {sym for rule in rules for sym in rule.expansion} | {rule.origin for rule in rules}

    # foreach grammar rule X ::= Y(1) ... Y(k)
    # if k=0 or {Y(1),...,Y(k)} subset of NULLABLE then
    #   NULLABLE = NULLABLE union {X}
    # for i = 1 to k
    #   if i=1 or {Y(1),...,Y(i-1)} subset of NULLABLE then
    #     FIRST(X) = FIRST(X) union FIRST(Y(i))
    #   for j = i+1 to k
    #     if i=k or {Y(i+1),...Y(k)} subset of NULLABLE then
    #       FOLLOW(Y(i)) = FOLLOW(Y(i)) union FOLLOW(X)
    #     if i+1=j or {Y(i+1),...,Y(j-1)} subset of NULLABLE then
    #       FOLLOW(Y(i)) = FOLLOW(Y(i)) union FIRST(Y(j))
    # until none of NULLABLE,FIRST,FOLLOW changed in last iteration

    NULLABLE = set()
    FIRST = {}
    FOLLOW = {}
    for sym in symbols:
        FIRST[sym]={sym} if sym.is_term else set()
        FOLLOW[sym]=set()

    # Calculate NULLABLE and FIRST
    changed = True
    while changed:
        changed = False

        for rule in rules:
            if set(rule.expansion) <= NULLABLE:
                if update_set(NULLABLE, {rule.origin}):
                    changed = True

            for i, sym in enumerate(rule.expansion):
                if set(rule.expansion[:i]) <= NULLABLE:
                    if update_set(FIRST[rule.origin], FIRST[sym]):
                        changed = True
                else:
                    break

    # Calculate FOLLOW
    changed = True
    while changed:
        changed = False

        for rule in rules:
            for i, sym in enumerate(rule.expansion):
                if i==len(rule.expansion)-1 or set(rule.expansion[i+1:]) <= NULLABLE:
                    if update_set(FOLLOW[sym], FOLLOW[rule.origin]):
                        changed = True

                for j in range(i+1, len(rule.expansion)):
                    if set(rule.expansion[i+1:j]) <= NULLABLE:
                        if update_set(FOLLOW[sym], FIRST[rule.expansion[j]]):
                            changed = True

    return FIRST, FOLLOW, NULLABLE


class GrammarAnalyzer:
    def __init__(self, parser_conf: ParserConf, debug: bool=False, strict: bool=False):
        self.debug = debug
        self.strict = strict

        root_rules = {start: Rule(NonTerminal('$root_' + start), [NonTerminal(start), Terminal('$END')])
                      for start in parser_conf.start}

        rules = parser_conf.rules + list(root_rules.values())
        self.rules_by_origin: Dict[NonTerminal, List[Rule]] = classify(rules, lambda r: r.origin)

        if len(rules) != len(set(rules)):
            duplicates = [item for item, count in Counter(rules).items() if count > 1]
            raise GrammarError("Rules defined twice: %s" % ', '.join(str(i) for i in duplicates))

        for r in rules:
            for sym in r.expansion:
                if not (sym.is_term or sym in self.rules_by_origin):
                    raise GrammarError("Using an undefined rule: %s" % sym)

        self.start_states = {start: self.expand_rule(root_rule.origin)
                             for start, root_rule in root_rules.items()}

        self.end_states = {start: fzset({RulePtr(root_rule, len(root_rule.expansion))})
                           for start, root_rule in root_rules.items()}

        lr0_root_rules = {start: Rule(NonTerminal('$root_' + start), [NonTerminal(start)])
                for start in parser_conf.start}

        lr0_rules = parser_conf.rules + list(lr0_root_rules.values())
        assert(len(lr0_rules) == len(set(lr0_rules)))

        self.lr0_rules_by_origin = classify(lr0_rules, lambda r: r.origin)

        # cache RulePtr(r, 0) in r (no duplicate RulePtr objects)
        self.lr0_start_states = {start: LR0ItemSet([RulePtr(root_rule, 0)], self.expand_rule(root_rule.origin, self.lr0_rules_by_origin))
                for start, root_rule in lr0_root_rules.items()}

        self.FIRST, self.FOLLOW, self.NULLABLE = calculate_sets(rules)

    def expand_rule(self, source_rule: NonTerminal, rules_by_origin=None) -> OrderedSet[RulePtr]:
        "Returns all init_ptrs accessible by rule (recursive)"

        if rules_by_origin is None:
            rules_by_origin = self.rules_by_origin

        init_ptrs = OrderedSet[RulePtr]()
        def _expand_rule(rule: NonTerminal) -> Iterator[NonTerminal]:
            assert not rule.is_term, rule

            for r in rules_by_origin[rule]:
                init_ptr = RulePtr(r, 0)
                init_ptrs.add(init_ptr)

                if r.expansion: # if not empty rule
                    new_r = init_ptr.next
                    if not new_r.is_term:
                        assert isinstance(new_r, NonTerminal)
                        yield new_r

        for _ in bfs([source_rule], _expand_rule):
            pass

        return init_ptrs



from __future__ import annotations

from pathlib import Path

import pytest

from poetry.core.masonry.utils.package_include import PackageInclude


fixtures_dir = Path(__file__).parent / "fixtures"
with_includes = fixtures_dir / "with_includes"


def test_package_include_with_multiple_dirs() -> None:
    pkg_include = PackageInclude(base=fixtures_dir, include="with_includes", formats=[])
    assert pkg_include.elements == [
        with_includes / "__init__.py",
        with_includes / "bar",
        with_includes / "bar/baz.py",
        with_includes / "extra_package",
        with_includes / "extra_package/some_dir",
        with_includes / "extra_package/some_dir/foo.py",
        with_includes / "extra_package/some_dir/quux.py",
        with_includes / "not_a_python_pkg",
        with_includes / "not_a_python_pkg/baz.txt",
    ]


def test_package_include_with_simple_dir() -> None:
    pkg_include = PackageInclude(base=with_includes, include="bar", formats=[])
    assert pkg_include.elements == [with_includes / "bar/baz.py"]


def test_package_include_with_nested_dir() -> None:
    pkg_include = PackageInclude(
        base=with_includes, include="extra_package/**/*.py", formats=[]
    )
    assert pkg_include.elements == [
        with_includes / "extra_package/some_dir/foo.py",
        with_includes / "extra_package/some_dir/quux.py",
    ]


def test_package_include_with_no_python_files_in_dir() -> None:
    with pytest.raises(ValueError) as e:
        PackageInclude(base=with_includes, include="not_a_python_pkg", formats=[])

    assert str(e.value) == "not_a_python_pkg is not a package."


def test_package_include_with_non_existent_directory() -> None:
    with pytest.raises(ValueError) as e:
        PackageInclude(base=with_includes, include="not_a_dir", formats=[])

    err_str = str(with_includes / "not_a_dir") + " does not contain any element"

    assert str(e.value) == err_str


def test_pep_561_stub_only_package_good_name_suffix() -> None:
    pkg_include = PackageInclude(
        base=fixtures_dir / "pep_561_stub_only", include="good-stubs", formats=[]
    )
    assert pkg_include.elements == [
        fixtures_dir / "pep_561_stub_only/good-stubs/__init__.pyi",
        fixtures_dir / "pep_561_stub_only/good-stubs/module.pyi",
    ]


def test_pep_561_stub_only_partial_namespace_package_good_name_suffix() -> None:
    pkg_include = PackageInclude(
        base=fixtures_dir / "pep_561_stub_only_partial_namespace",
        include="good-stubs",
        formats=[],
    )
    assert pkg_include.elements == [
        fixtures_dir / "pep_561_stub_only_partial_namespace/good-stubs/module.pyi",
        fixtures_dir / "pep_561_stub_only_partial_namespace/good-stubs/subpkg/",
        fixtures_dir
        / "pep_561_stub_only_partial_namespace/good-stubs/subpkg/__init__.pyi",
        fixtures_dir / "pep_561_stub_only_partial_namespace/good-stubs/subpkg/py.typed",
    ]


def test_pep_561_stub_only_package_bad_name_suffix() -> None:
    with pytest.raises(ValueError) as e:
        PackageInclude(
            base=fixtures_dir / "pep_561_stub_only", include="bad", formats=[]
        )

    assert str(e.value) == "bad is not a package."



from __future__ import annotations

import unicodedata

from pathlib import Path

from packaging.utils import canonicalize_name


def combine_unicode(string: str) -> str:
    return unicodedata.normalize("NFC", string)


def module_name(name: str) -> str:
    return canonicalize_name(name).replace("-", "_")


def parse_requires(requires: str) -> list[str]:
    lines = requires.split("\n")

    requires_dist = []
    current_marker = None
    for line in lines:
        line = line.strip()
        if not line:
            continue

        if line.startswith("["):
            # extras or conditional dependencies
            marker = line.lstrip("[").rstrip("]")
            if ":" not in marker:
                extra, marker = marker, ""
            else:
                extra, marker = marker.split(":")

            if extra:
                if marker:
                    marker = f'{marker} and extra == "{extra}"'
                else:
                    marker = f'extra == "{extra}"'

            if marker:
                current_marker = marker

            continue

        if current_marker:
            line = f"{line} ; {current_marker}"

        requires_dist.append(line)

    return requires_dist


def readme_content_type(path: str | Path) -> str:
    suffix = Path(path).suffix
    if suffix == ".rst":
        return "text/x-rst"
    elif suffix in (".md", ".markdown"):
        return "text/markdown"
    else:
        return "text/plain"



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from poetry.core.constraints.generic import AnyConstraint
from poetry.core.constraints.generic import Constraint
from poetry.core.constraints.generic import EmptyConstraint
from poetry.core.constraints.generic import MultiConstraint
from poetry.core.constraints.generic import UnionConstraint
from poetry.core.constraints.generic.constraint import ExtraConstraint
from poetry.core.constraints.generic.multi_constraint import ExtraMultiConstraint


if TYPE_CHECKING:
    from poetry.core.constraints.generic import BaseConstraint


@pytest.mark.parametrize(
    ("constraint1", "constraint2", "expected"),
    [
        (Constraint("win32"), Constraint("win32"), True),
        (Constraint("win32"), Constraint("linux"), False),
        (Constraint("win32", "!="), Constraint("win32"), False),
        (Constraint("win32", "!="), Constraint("linux"), True),
        (Constraint("tegra", "in"), Constraint("1.2-tegra"), True),
        (Constraint("tegra", "in"), Constraint("1.2-teg"), False),
        (Constraint("tegra", "not in"), Constraint("1.2-tegra"), False),
        (Constraint("tegra", "not in"), Constraint("1.2-teg"), True),
    ],
)
def test_allows(
    constraint1: Constraint, constraint2: Constraint, expected: bool
) -> None:
    assert constraint1.allows(constraint2) is expected
    # allows_any() and allows_all() should be the same as allows()
    # if the second constraint is a `==` constraint
    assert constraint1.allows_any(constraint2) is expected
    assert constraint1.allows_all(constraint2) is expected


@pytest.mark.parametrize(
    ("constraint1", "constraint2", "expected_any", "expected_all"),
    [
        (Constraint("win32"), EmptyConstraint(), False, True),
        (Constraint("win32"), AnyConstraint(), True, False),
        (Constraint("win32"), Constraint("win32"), True, True),
        (Constraint("win32"), Constraint("linux"), False, False),
        (Constraint("win32"), Constraint("win32", "!="), False, False),
        (Constraint("win32"), Constraint("linux", "!="), True, False),
        (
            Constraint("win32"),
            UnionConstraint(Constraint("win32"), Constraint("linux")),
            True,
            False,
        ),
        (
            Constraint("win32"),
            UnionConstraint(Constraint("darwin"), Constraint("linux")),
            False,
            False,
        ),
        (
            Constraint("win32"),
            UnionConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
            True,
            False,
        ),
        (
            Constraint("win32"),
            UnionConstraint(Constraint("darwin", "!="), Constraint("linux", "!=")),
            True,
            False,
        ),
        (
            Constraint("win32"),
            MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
            False,
            False,
        ),
        (
            Constraint("win32"),
            MultiConstraint(Constraint("darwin", "!="), Constraint("linux", "!=")),
            True,
            False,
        ),
        (Constraint("win32", "!="), EmptyConstraint(), False, True),
        (Constraint("win32", "!="), AnyConstraint(), True, False),
        (Constraint("win32", "!="), Constraint("win32"), False, False),
        (Constraint("win32", "!="), Constraint("linux"), True, True),
        (Constraint("win32", "!="), Constraint("win32", "!="), True, True),
        (Constraint("win32", "!="), Constraint("linux", "!="), True, False),
        (
            Constraint("win32", "!="),
            UnionConstraint(Constraint("win32"), Constraint("linux")),
            True,
            False,
        ),
        (
            Constraint("win32", "!="),
            UnionConstraint(Constraint("darwin"), Constraint("linux")),
            True,
            True,
        ),
        (
            Constraint("win32", "!="),
            UnionConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
            True,
            False,
        ),
        (
            Constraint("win32", "!="),
            UnionConstraint(Constraint("darwin", "!="), Constraint("linux", "!=")),
            True,
            False,
        ),
        (
            Constraint("win32", "!="),
            MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
            True,
            True,
        ),
        (
            Constraint("win32", "!="),
            MultiConstraint(Constraint("darwin", "!="), Constraint("linux", "!=")),
            True,
            False,
        ),
        (
            Constraint("tegra", "not in"),
            Constraint("tegra", "not in"),
            True,
            True,
        ),
        (
            Constraint("tegra", "in"),
            Constraint("tegra", "not in"),
            False,
            False,
        ),
        (
            Constraint("tegra", "in"),
            Constraint("tegra", "in"),
            True,
            True,
        ),
        (
            Constraint("tegra", "not in"),
            Constraint("tegra", "in"),
            False,
            False,
        ),
        (
            Constraint("tegra", "in"),
            Constraint("teg", "in"),
            True,
            False,
        ),
        (
            Constraint("teg", "in"),
            Constraint("tegra", "in"),
            True,
            True,
        ),
        (
            Constraint("teg", "not in"),
            Constraint("tegra", "not in"),
            True,
            False,
        ),
        (
            Constraint("tegra", "not in"),
            Constraint("teg", "not in"),
            True,
            True,
        ),
        (
            Constraint("teg", "not in"),
            Constraint("tegra", "in"),
            False,
            False,
        ),
        (
            Constraint("tegra", "in"),
            Constraint("teg", "not in"),
            False,
            False,
        ),
        (
            Constraint("teg", "in"),
            Constraint("tegra", "not in"),
            True,
            False,
        ),
        (
            Constraint("tegra", "not in"),
            Constraint("teg", "in"),
            True,
            False,
        ),
        (
            Constraint("tegra", "in"),
            Constraint("rpi", "in"),
            True,
            False,
        ),
        (
            Constraint("1.2.3-tegra", "!="),
            Constraint("tegra", "in"),
            True,
            False,
        ),
        (
            Constraint("tegra", "in"),
            Constraint("1.2.3-tegra", "!="),
            True,
            False,
        ),
        (
            Constraint("1.2.3-tegra", "!="),
            Constraint("teg", "in"),
            True,
            False,
        ),
        (
            Constraint("teg", "in"),
            Constraint("1.2.3-tegra", "!="),
            True,
            False,
        ),
        (
            Constraint("1.2.3-tegra", "!="),
            Constraint("tegra", "not in"),
            True,
            True,
        ),
        (
            Constraint("tegra", "not in"),
            Constraint("1.2.3-tegra", "!="),
            True,
            False,
        ),
        (
            Constraint("1.2.3-tegra", "!="),
            Constraint("teg", "not in"),
            True,
            True,
        ),
        (
            Constraint("teg", "not in"),
            Constraint("1.2.3-tegra", "!="),
            True,
            False,
        ),
        (
            Constraint("1.2.3-tegra", "=="),
            Constraint("tegra", "in"),
            True,
            False,
        ),
        (
            Constraint("1.2.3-tegra", "=="),
            Constraint("tegra", "not in"),
            False,
            False,
        ),
    ],
)
def test_allows_any_and_allows_all(
    constraint1: Constraint,
    constraint2: BaseConstraint,
    expected_any: bool,
    expected_all: bool,
) -> None:
    assert constraint1.allows_any(constraint2) is expected_any
    assert constraint1.allows_all(constraint2) is expected_all


@pytest.mark.parametrize(
    ("constraint", "inverted"),
    [
        (EmptyConstraint(), AnyConstraint()),
        (Constraint("foo"), Constraint("foo", "!=")),
        (
            MultiConstraint(Constraint("foo", "!="), Constraint("bar", "!=")),
            UnionConstraint(Constraint("foo"), Constraint("bar")),
        ),
        (Constraint("tegra", "not in"), Constraint("tegra", "in")),
    ],
)
def test_invert(constraint: BaseConstraint, inverted: BaseConstraint) -> None:
    assert constraint.invert() == inverted
    assert inverted.invert() == constraint


@pytest.mark.parametrize(
    ("constraint", "inverted"),
    [
        (ExtraConstraint("foo"), ExtraConstraint("foo", "!=")),
        (
            ExtraMultiConstraint(ExtraConstraint("foo"), ExtraConstraint("bar", "!=")),
            UnionConstraint(ExtraConstraint("foo", "!="), ExtraConstraint("bar")),
        ),
    ],
)
def test_invert_extra(constraint: BaseConstraint, inverted: BaseConstraint) -> None:
    assert constraint.invert() == inverted
    assert inverted.invert() == constraint


@pytest.mark.parametrize(
    ("constraint1", "constraint2", "expected"),
    [
        (
            EmptyConstraint(),
            Constraint("win32"),
            EmptyConstraint(),
        ),
        (
            EmptyConstraint(),
            MultiConstraint(Constraint("darwin", "!="), Constraint("linux", "!=")),
            EmptyConstraint(),
        ),
        (
            EmptyConstraint(),
            UnionConstraint(Constraint("win32"), Constraint("linux")),
            EmptyConstraint(),
        ),
        (
            AnyConstraint(),
            Constraint("win32"),
            Constraint("win32"),
        ),
        (
            AnyConstraint(),
            MultiConstraint(Constraint("darwin", "!="), Constraint("linux", "!=")),
            MultiConstraint(Constraint("darwin", "!="), Constraint("linux", "!=")),
        ),
        (
            AnyConstraint(),
            UnionConstraint(Constraint("win32"), Constraint("linux")),
            UnionConstraint(Constraint("win32"), Constraint("linux")),
        ),
        (
            EmptyConstraint(),
            AnyConstraint(),
            EmptyConstraint(),
        ),
        (
            EmptyConstraint(),
            EmptyConstraint(),
            EmptyConstraint(),
        ),
        (
            AnyConstraint(),
            AnyConstraint(),
            AnyConstraint(),
        ),
        (
            Constraint("win32"),
            Constraint("win32"),
            Constraint("win32"),
        ),
        (
            Constraint("win32"),
            Constraint("linux"),
            EmptyConstraint(),
        ),
        (
            Constraint("win32"),
            MultiConstraint(Constraint("darwin", "!="), Constraint("linux", "!=")),
            Constraint("win32"),
        ),
        (
            Constraint("win32"),
            MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
            EmptyConstraint(),
        ),
        (
            Constraint("win32"),
            UnionConstraint(Constraint("win32"), Constraint("linux")),
            Constraint("win32"),
        ),
        (
            Constraint("win32"),
            UnionConstraint(Constraint("linux"), Constraint("linux2")),
            EmptyConstraint(),
        ),
        (
            Constraint("win32"),
            Constraint("linux", "!="),
            Constraint("win32"),
        ),
        (
            Constraint("win32", "!="),
            Constraint("linux"),
            Constraint("linux"),
        ),
        (
            Constraint("win32", "!="),
            Constraint("linux", "!="),
            (
                MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
                MultiConstraint(Constraint("linux", "!="), Constraint("win32", "!=")),
            ),
        ),
        (
            Constraint("win32", "!="),
            MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
            MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
        ),
        (
            Constraint("darwin", "!="),
            MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
            MultiConstraint(
                Constraint("win32", "!="),
                Constraint("linux", "!="),
                Constraint("darwin", "!="),
            ),
        ),
        (
            Constraint("win32", "!="),
            UnionConstraint(Constraint("win32"), Constraint("linux")),
            Constraint("linux"),
        ),
        (
            Constraint("win32", "!="),
            UnionConstraint(
                Constraint("win32"), Constraint("linux"), Constraint("darwin")
            ),
            UnionConstraint(Constraint("linux"), Constraint("darwin")),
        ),
        (
            Constraint("win32", "!="),
            UnionConstraint(Constraint("linux"), Constraint("linux2")),
            UnionConstraint(Constraint("linux"), Constraint("linux2")),
        ),
        (
            UnionConstraint(Constraint("win32"), Constraint("linux")),
            UnionConstraint(Constraint("win32"), Constraint("linux")),
            UnionConstraint(Constraint("win32"), Constraint("linux")),
        ),
        (
            UnionConstraint(Constraint("win32"), Constraint("linux")),
            UnionConstraint(Constraint("linux"), Constraint("win32")),
            (
                UnionConstraint(Constraint("win32"), Constraint("linux")),
                UnionConstraint(Constraint("linux"), Constraint("win32")),
            ),
        ),
        (
            UnionConstraint(Constraint("win32"), Constraint("linux")),
            UnionConstraint(Constraint("win32"), Constraint("darwin")),
            Constraint("win32"),
        ),
        (
            UnionConstraint(
                Constraint("win32"), Constraint("linux"), Constraint("darwin")
            ),
            UnionConstraint(
                Constraint("win32"), Constraint("cygwin"), Constraint("darwin")
            ),
            UnionConstraint(
                Constraint("win32"),
                Constraint("darwin"),
            ),
        ),
        (
            UnionConstraint(Constraint("win32"), Constraint("linux")),
            MultiConstraint(Constraint("win32", "!="), Constraint("darwin", "!=")),
            Constraint("linux"),
        ),
        (
            UnionConstraint(Constraint("win32"), Constraint("linux")),
            MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
            EmptyConstraint(),
        ),
        (
            MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
            MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
            MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
        ),
        (
            MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
            MultiConstraint(Constraint("linux", "!="), Constraint("win32", "!=")),
            (
                MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
                MultiConstraint(Constraint("linux", "!="), Constraint("win32", "!=")),
            ),
        ),
        (
            MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
            MultiConstraint(Constraint("win32", "!="), Constraint("darwin", "!=")),
            (
                MultiConstraint(
                    Constraint("win32", "!="),
                    Constraint("linux", "!="),
                    Constraint("darwin", "!="),
                ),
                MultiConstraint(
                    Constraint("win32", "!="),
                    Constraint("darwin", "!="),
                    Constraint("linux", "!="),
                ),
            ),
        ),
        (
            Constraint("tegra", "not in"),
            Constraint("tegra", "not in"),
            Constraint("tegra", "not in"),
        ),
        (
            Constraint("tegra", "in"),
            Constraint("tegra", "not in"),
            EmptyConstraint(),
        ),
        (
            Constraint("tegra", "in"),
            Constraint("tegra", "in"),
            Constraint("tegra", "in"),
        ),
        (
            Constraint("teg", "in"),
            Constraint("tegra", "in"),
            Constraint("tegra", "in"),
        ),
        (
            Constraint("teg", "not in"),
            Constraint("tegra", "in"),
            EmptyConstraint(),
        ),
        (
            Constraint("teg", "in"),
            Constraint("tegra", "not in"),
            (
                MultiConstraint(Constraint("teg", "in"), Constraint("tegra", "not in")),
                MultiConstraint(Constraint("tegra", "not in"), Constraint("teg", "in")),
            ),
        ),
        (
            Constraint("tegra", "in"),
            Constraint("rpi", "in"),
            (
                MultiConstraint(Constraint("tegra", "in"), Constraint("rpi", "in")),
                MultiConstraint(Constraint("rpi", "in"), Constraint("tegra", "in")),
            ),
        ),
        (
            Constraint("tegra", "in"),
            Constraint("1.2.3-tegra", "=="),
            Constraint("1.2.3-tegra", "=="),
        ),
        (
            Constraint("tegra", "in"),
            Constraint("1.2.3-tegra", "!="),
            (
                MultiConstraint(
                    Constraint("tegra", "in"), Constraint("1.2.3-tegra", "!=")
                ),
                MultiConstraint(
                    Constraint("1.2.3-tegra", "!="),
                    Constraint("tegra", "in"),
                ),
            ),
        ),
        (
            Constraint("tegra", "not in"),
            Constraint("1.2.3-tegra", "=="),
            EmptyConstraint(),
        ),
        (
            Constraint("tegra", "not in"),
            Constraint("1.2.3-tegra", "!="),
            Constraint("tegra", "not in"),
        ),
        (
            Constraint("tegra", "not in"),
            Constraint("rpi", "not in"),
            (
                MultiConstraint(
                    Constraint("tegra", "not in"),
                    Constraint("rpi", "not in"),
                ),
                MultiConstraint(
                    Constraint("rpi", "not in"),
                    Constraint("tegra", "not in"),
                ),
            ),
        ),
    ],
)
def test_intersect(
    constraint1: BaseConstraint,
    constraint2: BaseConstraint,
    expected: BaseConstraint | tuple[BaseConstraint, BaseConstraint],
) -> None:
    if not isinstance(expected, tuple):
        expected = (expected, expected)
    assert constraint1.intersect(constraint2) == expected[0]
    assert constraint2.intersect(constraint1) == expected[1]


@pytest.mark.parametrize(
    ("constraint1", "constraint2", "expected"),
    [
        (
            EmptyConstraint(),
            ExtraConstraint("extra1"),
            EmptyConstraint(),
        ),
        (
            EmptyConstraint(),
            ExtraMultiConstraint(
                ExtraConstraint("extra1", "!="), ExtraConstraint("extra2", "!=")
            ),
            EmptyConstraint(),
        ),
        (
            EmptyConstraint(),
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            EmptyConstraint(),
        ),
        (
            AnyConstraint(),
            ExtraConstraint("extra1"),
            ExtraConstraint("extra1"),
        ),
        (
            AnyConstraint(),
            ExtraMultiConstraint(
                ExtraConstraint("extra1", "!="), ExtraConstraint("extra2", "!=")
            ),
            ExtraMultiConstraint(
                ExtraConstraint("extra1", "!="), ExtraConstraint("extra2", "!=")
            ),
        ),
        (
            AnyConstraint(),
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
        ),
        (
            ExtraConstraint("extra1"),
            ExtraConstraint("extra1"),
            ExtraConstraint("extra1"),
        ),
        (
            ExtraConstraint("extra1"),
            ExtraConstraint("extra1", "!="),
            EmptyConstraint(),
        ),
        (
            ExtraConstraint("extra1"),
            ExtraConstraint("extra2"),
            (
                ExtraMultiConstraint(
                    ExtraConstraint("extra1"), ExtraConstraint("extra2")
                ),
                ExtraMultiConstraint(
                    ExtraConstraint("extra2"), ExtraConstraint("extra1")
                ),
            ),
        ),
        (
            ExtraConstraint("extra1"),
            ExtraMultiConstraint(
                ExtraConstraint("extra2", "!="), ExtraConstraint("extra3", "!=")
            ),
            ExtraMultiConstraint(
                ExtraConstraint("extra2", "!="),
                ExtraConstraint("extra3", "!="),
                ExtraConstraint("extra1"),
            ),
        ),
        (
            ExtraConstraint("extra1"),
            ExtraMultiConstraint(
                ExtraConstraint("extra1", "!="), ExtraConstraint("extra2", "!=")
            ),
            EmptyConstraint(),
        ),
        (
            ExtraConstraint("extra1"),
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            ExtraConstraint("extra1"),
        ),
        (
            ExtraConstraint("extra1"),
            UnionConstraint(ExtraConstraint("extra2"), ExtraConstraint("extra3")),
            UnionConstraint(
                ExtraMultiConstraint(
                    ExtraConstraint("extra2"), ExtraConstraint("extra1")
                ),
                ExtraMultiConstraint(
                    ExtraConstraint("extra3"), ExtraConstraint("extra1")
                ),
            ),
        ),
        (
            ExtraConstraint("extra1"),
            ExtraConstraint("extra2", "!="),
            (
                ExtraMultiConstraint(
                    ExtraConstraint("extra1"), ExtraConstraint("extra2", "!=")
                ),
                ExtraMultiConstraint(
                    ExtraConstraint("extra2", "!="), ExtraConstraint("extra1")
                ),
            ),
        ),
        (
            ExtraConstraint("extra1", "!="),
            ExtraConstraint("extra2"),
            (
                ExtraMultiConstraint(
                    ExtraConstraint("extra1", "!="), ExtraConstraint("extra2")
                ),
                ExtraMultiConstraint(
                    ExtraConstraint("extra2"), ExtraConstraint("extra1", "!=")
                ),
            ),
        ),
        (
            ExtraConstraint("extra1", "!="),
            ExtraConstraint("extra2", "!="),
            (
                ExtraMultiConstraint(
                    ExtraConstraint("extra1", "!="), ExtraConstraint("extra2", "!=")
                ),
                ExtraMultiConstraint(
                    ExtraConstraint("extra2", "!="), ExtraConstraint("extra1", "!=")
                ),
            ),
        ),
        (
            ExtraConstraint("extra1", "!="),
            ExtraMultiConstraint(
                ExtraConstraint("extra1", "!="), ExtraConstraint("extra2", "!=")
            ),
            ExtraMultiConstraint(
                ExtraConstraint("extra1", "!="), ExtraConstraint("extra2", "!=")
            ),
        ),
        (
            ExtraConstraint("extra3", "!="),
            ExtraMultiConstraint(
                ExtraConstraint("extra1", "!="), ExtraConstraint("extra2", "!=")
            ),
            ExtraMultiConstraint(
                ExtraConstraint("extra1", "!="),
                ExtraConstraint("extra2", "!="),
                ExtraConstraint("extra3", "!="),
            ),
        ),
        (
            ExtraConstraint("extra1", "!="),
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            ExtraMultiConstraint(
                ExtraConstraint("extra2"), ExtraConstraint("extra1", "!=")
            ),
        ),
        (
            ExtraConstraint("extra1", "!="),
            UnionConstraint(
                ExtraConstraint("extra1"),
                ExtraConstraint("extra2"),
                ExtraConstraint("extra3"),
            ),
            UnionConstraint(
                ExtraMultiConstraint(
                    ExtraConstraint("extra2"), ExtraConstraint("extra1", "!=")
                ),
                ExtraMultiConstraint(
                    ExtraConstraint("extra3"), ExtraConstraint("extra1", "!=")
                ),
            ),
        ),
        (
            ExtraConstraint("extra1", "!="),
            UnionConstraint(ExtraConstraint("extra2"), ExtraConstraint("extra3")),
            UnionConstraint(
                ExtraMultiConstraint(
                    ExtraConstraint("extra2"), ExtraConstraint("extra1", "!=")
                ),
                ExtraMultiConstraint(
                    ExtraConstraint("extra3"), ExtraConstraint("extra1", "!=")
                ),
            ),
        ),
        (
            ExtraMultiConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            ExtraMultiConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            ExtraMultiConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
        ),
        (
            ExtraMultiConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            ExtraMultiConstraint(ExtraConstraint("extra2"), ExtraConstraint("extra1")),
            (
                ExtraMultiConstraint(
                    ExtraConstraint("extra1"), ExtraConstraint("extra2")
                ),
                ExtraMultiConstraint(
                    ExtraConstraint("extra2"), ExtraConstraint("extra1")
                ),
            ),
        ),
        (
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
        ),
        (
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            UnionConstraint(
                ExtraConstraint("extra1"),
                ExtraConstraint("extra2"),
                ExtraConstraint("extra3"),
            ),
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
        ),
        (
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            UnionConstraint(ExtraConstraint("extra2"), ExtraConstraint("extra1")),
            (
                UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
                UnionConstraint(ExtraConstraint("extra2"), ExtraConstraint("extra1")),
            ),
        ),
        (
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra3")),
            (
                UnionConstraint(
                    ExtraConstraint("extra1"),
                    ExtraMultiConstraint(
                        ExtraConstraint("extra1"), ExtraConstraint("extra3")
                    ),
                    ExtraMultiConstraint(
                        ExtraConstraint("extra2"), ExtraConstraint("extra1")
                    ),
                    ExtraMultiConstraint(
                        ExtraConstraint("extra2"), ExtraConstraint("extra3")
                    ),
                ),
                UnionConstraint(
                    ExtraConstraint("extra1"),
                    ExtraMultiConstraint(
                        ExtraConstraint("extra1"), ExtraConstraint("extra2")
                    ),
                    ExtraMultiConstraint(
                        ExtraConstraint("extra3"), ExtraConstraint("extra1")
                    ),
                    ExtraMultiConstraint(
                        ExtraConstraint("extra3"), ExtraConstraint("extra2")
                    ),
                ),
            ),
        ),
        (
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            ExtraMultiConstraint(
                ExtraConstraint("extra1", "!="), ExtraConstraint("extra3", "!=")
            ),
            ExtraMultiConstraint(
                ExtraConstraint("extra2"),
                ExtraConstraint("extra1", "!="),
                ExtraConstraint("extra3", "!="),
            ),
        ),
        (
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            ExtraMultiConstraint(
                ExtraConstraint("extra1", "!="), ExtraConstraint("extra2", "!=")
            ),
            EmptyConstraint(),
        ),
        (
            ExtraMultiConstraint(
                ExtraConstraint("extra1", "!="), ExtraConstraint("extra2", "!=")
            ),
            ExtraMultiConstraint(
                ExtraConstraint("extra1", "!="), ExtraConstraint("extra3", "!=")
            ),
            (
                ExtraMultiConstraint(
                    ExtraConstraint("extra1", "!="),
                    ExtraConstraint("extra2", "!="),
                    ExtraConstraint("extra3", "!="),
                ),
                ExtraMultiConstraint(
                    ExtraConstraint("extra1", "!="),
                    ExtraConstraint("extra3", "!="),
                    ExtraConstraint("extra2", "!="),
                ),
            ),
        ),
        (
            ExtraMultiConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            ExtraMultiConstraint(
                ExtraConstraint("extra1", "!="), ExtraConstraint("extra3", "!=")
            ),
            EmptyConstraint(),
        ),
        (
            ExtraMultiConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            ExtraMultiConstraint(
                ExtraConstraint("extra3", "!="), ExtraConstraint("extra4", "!=")
            ),
            (
                ExtraMultiConstraint(
                    ExtraConstraint("extra1"),
                    ExtraConstraint("extra2"),
                    ExtraConstraint("extra3", "!="),
                    ExtraConstraint("extra4", "!="),
                ),
                ExtraMultiConstraint(
                    ExtraConstraint("extra3", "!="),
                    ExtraConstraint("extra4", "!="),
                    ExtraConstraint("extra1"),
                    ExtraConstraint("extra2"),
                ),
            ),
        ),
        (
            ExtraMultiConstraint(
                ExtraConstraint("extra1", "!="), ExtraConstraint("extra2", "!=")
            ),
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            EmptyConstraint(),
        ),
        (
            ExtraMultiConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            ExtraMultiConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
        ),
    ],
)
def test_intersect_extra(
    constraint1: BaseConstraint,
    constraint2: BaseConstraint,
    expected: BaseConstraint | tuple[BaseConstraint, BaseConstraint],
) -> None:
    if not isinstance(expected, tuple):
        expected = (expected, expected)
    assert constraint1.intersect(constraint2) == expected[0]
    assert constraint2.intersect(constraint1) == expected[1]


@pytest.mark.parametrize(
    ("constraint1", "constraint2", "expected"),
    [
        (
            EmptyConstraint(),
            Constraint("win32"),
            Constraint("win32"),
        ),
        (
            EmptyConstraint(),
            MultiConstraint(Constraint("darwin", "!="), Constraint("linux", "!=")),
            MultiConstraint(Constraint("darwin", "!="), Constraint("linux", "!=")),
        ),
        (
            EmptyConstraint(),
            UnionConstraint(Constraint("win32"), Constraint("linux")),
            UnionConstraint(Constraint("win32"), Constraint("linux")),
        ),
        (
            AnyConstraint(),
            Constraint("win32"),
            AnyConstraint(),
        ),
        (
            AnyConstraint(),
            MultiConstraint(Constraint("darwin", "!="), Constraint("linux", "!=")),
            AnyConstraint(),
        ),
        (
            AnyConstraint(),
            UnionConstraint(Constraint("win32"), Constraint("linux")),
            AnyConstraint(),
        ),
        (
            EmptyConstraint(),
            AnyConstraint(),
            AnyConstraint(),
        ),
        (
            EmptyConstraint(),
            EmptyConstraint(),
            EmptyConstraint(),
        ),
        (
            AnyConstraint(),
            AnyConstraint(),
            AnyConstraint(),
        ),
        (
            Constraint("win32"),
            Constraint("win32"),
            Constraint("win32"),
        ),
        (
            Constraint("win32"),
            Constraint("linux"),
            (
                UnionConstraint(Constraint("win32"), Constraint("linux")),
                UnionConstraint(Constraint("linux"), Constraint("win32")),
            ),
        ),
        (
            Constraint("win32"),
            MultiConstraint(Constraint("darwin", "!="), Constraint("linux", "!=")),
            MultiConstraint(Constraint("darwin", "!="), Constraint("linux", "!=")),
        ),
        (
            Constraint("win32"),
            MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
            Constraint("linux", "!="),
        ),
        (
            Constraint("win32"),
            MultiConstraint(
                Constraint("win32", "!="),
                Constraint("linux", "!="),
                Constraint("darwin", "!="),
            ),
            MultiConstraint(Constraint("linux", "!="), Constraint("darwin", "!=")),
        ),
        (
            Constraint("win32"),
            UnionConstraint(Constraint("win32"), Constraint("linux")),
            UnionConstraint(Constraint("win32"), Constraint("linux")),
        ),
        (
            Constraint("win32"),
            UnionConstraint(Constraint("linux"), Constraint("linux2")),
            (
                UnionConstraint(
                    Constraint("win32"), Constraint("linux"), Constraint("linux2")
                ),
                UnionConstraint(
                    Constraint("linux"), Constraint("linux2"), Constraint("win32")
                ),
            ),
        ),
        (
            Constraint("win32"),
            Constraint("linux", "!="),
            Constraint("linux", "!="),
        ),
        (
            Constraint("win32", "!="),
            Constraint("linux"),
            Constraint("win32", "!="),
        ),
        (
            Constraint("win32", "!="),
            Constraint("linux", "!="),
            AnyConstraint(),
        ),
        (
            Constraint("win32", "!="),
            MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
            Constraint("win32", "!="),
        ),
        (
            Constraint("darwin", "!="),
            MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
            AnyConstraint(),
        ),
        (
            Constraint("win32", "!="),
            UnionConstraint(Constraint("win32"), Constraint("linux")),
            AnyConstraint(),
        ),
        (
            Constraint("win32", "!="),
            UnionConstraint(Constraint("linux"), Constraint("linux2")),
            Constraint("win32", "!="),
        ),
        (
            UnionConstraint(Constraint("win32"), Constraint("linux")),
            UnionConstraint(Constraint("win32"), Constraint("linux")),
            UnionConstraint(Constraint("win32"), Constraint("linux")),
        ),
        (
            UnionConstraint(Constraint("win32"), Constraint("linux")),
            UnionConstraint(Constraint("linux"), Constraint("win32")),
            (
                UnionConstraint(Constraint("win32"), Constraint("linux")),
                UnionConstraint(Constraint("linux"), Constraint("win32")),
            ),
        ),
        (
            UnionConstraint(Constraint("win32"), Constraint("linux")),
            UnionConstraint(Constraint("win32"), Constraint("darwin")),
            (
                UnionConstraint(
                    Constraint("win32"), Constraint("linux"), Constraint("darwin")
                ),
                UnionConstraint(
                    Constraint("win32"), Constraint("darwin"), Constraint("linux")
                ),
            ),
        ),
        (
            UnionConstraint(
                Constraint("win32"), Constraint("linux"), Constraint("darwin")
            ),
            UnionConstraint(
                Constraint("win32"), Constraint("cygwin"), Constraint("darwin")
            ),
            (
                UnionConstraint(
                    Constraint("win32"),
                    Constraint("linux"),
                    Constraint("darwin"),
                    Constraint("cygwin"),
                ),
                UnionConstraint(
                    Constraint("win32"),
                    Constraint("cygwin"),
                    Constraint("darwin"),
                    Constraint("linux"),
                ),
            ),
        ),
        (
            UnionConstraint(Constraint("win32"), Constraint("linux")),
            MultiConstraint(Constraint("win32", "!="), Constraint("darwin", "!=")),
            Constraint("darwin", "!="),
        ),
        (
            UnionConstraint(Constraint("win32"), Constraint("linux")),
            MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
            AnyConstraint(),
        ),
        (
            MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
            MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
            MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
        ),
        (
            MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
            MultiConstraint(Constraint("linux", "!="), Constraint("win32", "!=")),
            (
                MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
                MultiConstraint(Constraint("linux", "!="), Constraint("win32", "!=")),
            ),
        ),
        (
            MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
            MultiConstraint(Constraint("win32", "!="), Constraint("darwin", "!=")),
            Constraint("win32", "!="),
        ),
        (
            MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
            MultiConstraint(Constraint("bsd", "!="), Constraint("darwin", "!=")),
            AnyConstraint(),
        ),
        (
            Constraint("tegra", "not in"),
            Constraint("tegra", "not in"),
            Constraint("tegra", "not in"),
        ),
        (
            Constraint("tegra", "in"),
            Constraint("tegra", "not in"),
            AnyConstraint(),
        ),
        (
            Constraint("tegra", "in"),
            Constraint("tegra", "in"),
            Constraint("tegra", "in"),
        ),
        (
            Constraint("teg", "in"),
            Constraint("tegra", "in"),
            Constraint("teg", "in"),
        ),
        (
            Constraint("teg", "in"),
            Constraint("tegra", "not in"),
            AnyConstraint(),
        ),
        (
            Constraint("teg", "not in"),
            Constraint("tegra", "in"),
            (
                UnionConstraint(Constraint("teg", "not in"), Constraint("tegra", "in")),
                UnionConstraint(Constraint("tegra", "in"), Constraint("teg", "not in")),
            ),
        ),
        (
            Constraint("tegra", "in"),
            Constraint("rpi", "in"),
            (
                UnionConstraint(Constraint("tegra", "in"), Constraint("rpi", "in")),
                UnionConstraint(Constraint("rpi", "in"), Constraint("tegra", "in")),
            ),
        ),
        (
            Constraint("tegra", "not in"),
            Constraint("rpi", "not in"),
            AnyConstraint(),
        ),
        (
            Constraint("tegra", "in"),
            Constraint("1.2.3-tegra", "!="),
            AnyConstraint(),
        ),
        (
            Constraint("tegra", "not in"),
            Constraint("1.2.3-tegra", "!="),
            Constraint("1.2.3-tegra", "!="),
        ),
        (
            Constraint("tegra", "in"),
            Constraint("1.2.3-tegra", "=="),
            Constraint("tegra", "in"),
        ),
        (
            Constraint("tegra", "not in"),
            Constraint("1.2.3-tegra", "=="),
            (
                UnionConstraint(
                    Constraint("tegra", "not in"), Constraint("1.2.3-tegra", "==")
                ),
                UnionConstraint(
                    Constraint("1.2.3-tegra", "=="), Constraint("tegra", "not in")
                ),
            ),
        ),
    ],
)
def test_union(
    constraint1: BaseConstraint,
    constraint2: BaseConstraint,
    expected: BaseConstraint | tuple[BaseConstraint, BaseConstraint],
) -> None:
    if not isinstance(expected, tuple):
        expected = (expected, expected)

    assert constraint1.union(constraint2) == expected[0]
    assert constraint2.union(constraint1) == expected[1]


@pytest.mark.parametrize(
    ("constraint1", "constraint2", "expected"),
    [
        (
            EmptyConstraint(),
            ExtraConstraint("extra1"),
            ExtraConstraint("extra1"),
        ),
        (
            EmptyConstraint(),
            ExtraMultiConstraint(
                ExtraConstraint("extra1", "!="), ExtraConstraint("extra2", "!=")
            ),
            ExtraMultiConstraint(
                ExtraConstraint("extra1", "!="), ExtraConstraint("extra2", "!=")
            ),
        ),
        (
            EmptyConstraint(),
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
        ),
        (
            AnyConstraint(),
            ExtraConstraint("extra1"),
            AnyConstraint(),
        ),
        (
            AnyConstraint(),
            ExtraMultiConstraint(
                ExtraConstraint("extra1", "!="), ExtraConstraint("extra2", "!=")
            ),
            AnyConstraint(),
        ),
        (
            AnyConstraint(),
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            AnyConstraint(),
        ),
        (
            ExtraConstraint("extra1"),
            ExtraConstraint("extra1"),
            ExtraConstraint("extra1"),
        ),
        (
            ExtraConstraint("extra1"),
            ExtraConstraint("extra1", "!="),
            AnyConstraint(),
        ),
        (
            ExtraConstraint("extra1"),
            ExtraConstraint("extra2"),
            (
                UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
                UnionConstraint(ExtraConstraint("extra2"), ExtraConstraint("extra1")),
            ),
        ),
        (
            ExtraConstraint("extra1"),
            ExtraMultiConstraint(
                ExtraConstraint("extra2", "!="), ExtraConstraint("extra3", "!=")
            ),
            UnionConstraint(
                ExtraMultiConstraint(
                    ExtraConstraint("extra2", "!="), ExtraConstraint("extra3", "!=")
                ),
                ExtraConstraint("extra1"),
            ),
        ),
        (
            ExtraConstraint("extra1"),
            ExtraMultiConstraint(
                ExtraConstraint("extra1", "!="), ExtraConstraint("extra2", "!=")
            ),
            UnionConstraint(ExtraConstraint("extra2", "!="), ExtraConstraint("extra1")),
        ),
        (
            ExtraConstraint("extra1"),
            ExtraMultiConstraint(
                ExtraConstraint("extra1", "!="),
                ExtraConstraint("extra2", "!="),
                ExtraConstraint("extra3", "!="),
            ),
            UnionConstraint(
                ExtraMultiConstraint(
                    ExtraConstraint("extra1", "!="),
                    ExtraConstraint("extra2", "!="),
                    ExtraConstraint("extra3", "!="),
                ),
                ExtraConstraint("extra1"),
            ),
        ),
        (
            ExtraConstraint("extra1"),
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
        ),
        (
            ExtraConstraint("extra1"),
            UnionConstraint(ExtraConstraint("extra2"), ExtraConstraint("extra3")),
            (
                UnionConstraint(
                    ExtraConstraint("extra1"),
                    ExtraConstraint("extra2"),
                    ExtraConstraint("extra3"),
                ),
                UnionConstraint(
                    ExtraConstraint("extra2"),
                    ExtraConstraint("extra3"),
                    ExtraConstraint("extra1"),
                ),
            ),
        ),
        (
            ExtraConstraint("extra1"),
            ExtraConstraint("extra2", "!="),
            (
                UnionConstraint(
                    ExtraConstraint("extra1"), ExtraConstraint("extra2", "!=")
                ),
                UnionConstraint(
                    ExtraConstraint("extra2", "!="), ExtraConstraint("extra1")
                ),
            ),
        ),
        (
            ExtraConstraint("extra1", "!="),
            ExtraConstraint("extra2"),
            (
                UnionConstraint(
                    ExtraConstraint("extra1", "!="), ExtraConstraint("extra2")
                ),
                UnionConstraint(
                    ExtraConstraint("extra2"), ExtraConstraint("extra1", "!=")
                ),
            ),
        ),
        (
            ExtraConstraint("extra1", "!="),
            ExtraConstraint("extra2", "!="),
            (
                UnionConstraint(
                    ExtraConstraint("extra1", "!="), ExtraConstraint("extra2", "!=")
                ),
                UnionConstraint(
                    ExtraConstraint("extra2", "!="), ExtraConstraint("extra1", "!=")
                ),
            ),
        ),
        (
            ExtraConstraint("extra1", "!="),
            ExtraMultiConstraint(
                ExtraConstraint("extra1", "!="), ExtraConstraint("extra2", "!=")
            ),
            ExtraConstraint("extra1", "!="),
        ),
        (
            ExtraConstraint("extra1", "!="),
            ExtraMultiConstraint(
                ExtraConstraint("extra2", "!="), ExtraConstraint("extra3", "!=")
            ),
            UnionConstraint(
                ExtraMultiConstraint(
                    ExtraConstraint("extra2", "!="), ExtraConstraint("extra3", "!=")
                ),
                ExtraConstraint("extra1", "!="),
            ),
        ),
        (
            ExtraConstraint("extra1", "!="),
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            AnyConstraint(),
        ),
        (
            ExtraConstraint("extra1", "!="),
            UnionConstraint(ExtraConstraint("extra2"), ExtraConstraint("extra3")),
            (
                UnionConstraint(
                    ExtraConstraint("extra1", "!="),
                    ExtraConstraint("extra2"),
                    ExtraConstraint("extra3"),
                ),
                UnionConstraint(
                    ExtraConstraint("extra2"),
                    ExtraConstraint("extra3"),
                    ExtraConstraint("extra1", "!="),
                ),
            ),
        ),
        (
            ExtraMultiConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            ExtraMultiConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            ExtraMultiConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
        ),
        (
            ExtraMultiConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            ExtraMultiConstraint(ExtraConstraint("extra2"), ExtraConstraint("extra1")),
            (
                ExtraMultiConstraint(
                    ExtraConstraint("extra1"), ExtraConstraint("extra2")
                ),
                ExtraMultiConstraint(
                    ExtraConstraint("extra2"), ExtraConstraint("extra1")
                ),
            ),
        ),
        (
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
        ),
        (
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            UnionConstraint(ExtraConstraint("extra2"), ExtraConstraint("extra1")),
            (
                UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
                UnionConstraint(ExtraConstraint("extra2"), ExtraConstraint("extra1")),
            ),
        ),
        (
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra3")),
            (
                UnionConstraint(
                    ExtraConstraint("extra1"),
                    ExtraConstraint("extra2"),
                    ExtraConstraint("extra3"),
                ),
                UnionConstraint(
                    ExtraConstraint("extra1"),
                    ExtraConstraint("extra3"),
                    ExtraConstraint("extra2"),
                ),
            ),
        ),
        (
            UnionConstraint(
                ExtraConstraint("extra1"),
                ExtraConstraint("extra2"),
                ExtraConstraint("extra3"),
            ),
            UnionConstraint(
                ExtraConstraint("extra1"),
                ExtraConstraint("extra4"),
                ExtraConstraint("extra3"),
            ),
            (
                UnionConstraint(
                    ExtraConstraint("extra1"),
                    ExtraConstraint("extra2"),
                    ExtraConstraint("extra3"),
                    ExtraConstraint("extra4"),
                ),
                UnionConstraint(
                    ExtraConstraint("extra1"),
                    ExtraConstraint("extra4"),
                    ExtraConstraint("extra3"),
                    ExtraConstraint("extra2"),
                ),
            ),
        ),
        (
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            ExtraMultiConstraint(
                ExtraConstraint("extra3", "!="), ExtraConstraint("extra4", "!=")
            ),
            UnionConstraint(
                ExtraConstraint("extra1"),
                ExtraConstraint("extra2"),
                ExtraMultiConstraint(
                    ExtraConstraint("extra3", "!="), ExtraConstraint("extra4", "!=")
                ),
            ),
        ),
        (
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            ExtraMultiConstraint(
                ExtraConstraint("extra1", "!="), ExtraConstraint("extra3", "!=")
            ),
            UnionConstraint(
                ExtraConstraint("extra1"),
                ExtraConstraint("extra2"),
                ExtraConstraint("extra3", "!="),
            ),
        ),
        (
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            ExtraMultiConstraint(
                ExtraConstraint("extra1", "!="), ExtraConstraint("extra2", "!=")
            ),
            AnyConstraint(),
        ),
        (
            ExtraMultiConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            ExtraMultiConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            ExtraMultiConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
        ),
        (
            ExtraMultiConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            ExtraMultiConstraint(
                ExtraConstraint("extra1"),
                ExtraConstraint("extra2"),
                ExtraConstraint("extra3"),
            ),
            ExtraMultiConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
        ),
        (
            ExtraMultiConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            ExtraMultiConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra3")),
            (
                UnionConstraint(
                    ExtraMultiConstraint(
                        ExtraConstraint("extra1"), ExtraConstraint("extra2")
                    ),
                    ExtraMultiConstraint(
                        ExtraConstraint("extra1"), ExtraConstraint("extra3")
                    ),
                ),
                UnionConstraint(
                    ExtraMultiConstraint(
                        ExtraConstraint("extra1"), ExtraConstraint("extra3")
                    ),
                    ExtraMultiConstraint(
                        ExtraConstraint("extra1"), ExtraConstraint("extra2")
                    ),
                ),
            ),
        ),
        (
            ExtraMultiConstraint(
                ExtraConstraint("extra1", "!="), ExtraConstraint("extra2", "!=")
            ),
            ExtraMultiConstraint(
                ExtraConstraint("extra1", "!="), ExtraConstraint("extra3", "!=")
            ),
            (
                UnionConstraint(
                    ExtraMultiConstraint(
                        ExtraConstraint("extra1", "!="), ExtraConstraint("extra2", "!=")
                    ),
                    ExtraMultiConstraint(
                        ExtraConstraint("extra1", "!="), ExtraConstraint("extra3", "!=")
                    ),
                ),
                UnionConstraint(
                    ExtraMultiConstraint(
                        ExtraConstraint("extra1", "!="), ExtraConstraint("extra3", "!=")
                    ),
                    ExtraMultiConstraint(
                        ExtraConstraint("extra1", "!="), ExtraConstraint("extra2", "!=")
                    ),
                ),
            ),
        ),
        (
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
            ExtraMultiConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra3")),
            UnionConstraint(ExtraConstraint("extra1"), ExtraConstraint("extra2")),
        ),
    ],
)
def test_union_extra(
    constraint1: BaseConstraint,
    constraint2: BaseConstraint,
    expected: BaseConstraint | tuple[BaseConstraint, BaseConstraint],
) -> None:
    if not isinstance(expected, tuple):
        expected = (expected, expected)

    assert constraint1.union(constraint2) == expected[0]
    assert constraint2.union(constraint1) == expected[1]


def test_difference() -> None:
    c = Constraint("win32")

    assert c.difference(Constraint("win32")).is_empty()
    assert c.difference(Constraint("linux")) == c


@pytest.mark.parametrize(
    "constraint",
    [
        EmptyConstraint(),
        AnyConstraint(),
        Constraint("win32"),
        UnionConstraint(Constraint("win32"), Constraint("linux")),
        MultiConstraint(Constraint("win32", "!="), Constraint("linux", "!=")),
    ],
)
def test_constraints_are_hashable(constraint: BaseConstraint) -> None:
    # We're just testing that constraints are hashable, there's nothing much to say
    # about the result.
    hash(constraint)



from __future__ import annotations

import sys

from pathlib import Path
from typing import TYPE_CHECKING
from typing import cast

import pytest

from poetry.core.packages.dependency import Dependency
from poetry.core.packages.directory_dependency import DirectoryDependency


if TYPE_CHECKING:
    from pytest import LogCaptureFixture
    from pytest_mock import MockerFixture


DIST_PATH = Path(__file__).parent.parent / "fixtures" / "distributions"
SAMPLE_PROJECT = Path(__file__).parent.parent / "fixtures" / "sample_project"


def test_directory_dependency_does_not_exist(
    caplog: LogCaptureFixture, mocker: MockerFixture
) -> None:
    mock_exists = mocker.patch.object(Path, "exists")
    mock_exists.return_value = False
    dep = DirectoryDependency("demo", DIST_PATH / "invalid")
    assert len(caplog.records) == 1
    record = caplog.records[0]
    assert record.levelname == "WARNING"
    assert "does not exist" in record.message

    with pytest.raises(ValueError, match="does not exist"):
        dep.validate(raise_error=True)

    mock_exists.assert_called_once()


def test_directory_dependency_is_file(
    caplog: LogCaptureFixture, mocker: MockerFixture
) -> None:
    mock_is_file = mocker.patch.object(Path, "is_file")
    mock_is_file.return_value = True
    dep = DirectoryDependency("demo", DIST_PATH / "demo-0.1.0.tar.gz")
    assert len(caplog.records) == 1
    record = caplog.records[0]
    assert record.levelname == "WARNING"
    assert "is a file" in record.message

    with pytest.raises(ValueError, match="is a file"):
        dep.validate(raise_error=True)

    mock_is_file.assert_called_once()


def test_directory_dependency_is_not_a_python_project(
    caplog: LogCaptureFixture,
) -> None:
    dep = DirectoryDependency("demo", DIST_PATH)
    assert len(caplog.records) == 1
    record = caplog.records[0]
    assert record.levelname == "WARNING"
    assert "a Python package" in record.message

    with pytest.raises(ValueError, match=r"not .* a Python package"):
        dep.validate(raise_error=True)


def test_directory_dependency_minimal() -> None:
    path = Path(__file__).parent.parent / "fixtures" / "project_minimal"
    dep = DirectoryDependency("demo", path)
    dep.validate(raise_error=True)


def _test_directory_dependency_pep_508(
    name: str, path: Path, pep_508_input: str, pep_508_output: str | None = None
) -> None:
    dep = Dependency.create_from_pep_508(
        pep_508_input, relative_to=Path(__file__).parent
    )

    assert dep.is_directory()
    dep = cast("DirectoryDependency", dep)
    assert dep.name == name
    assert dep.path == path
    assert dep.to_pep_508() == (pep_508_output or pep_508_input)


def test_directory_dependency_pep_508_local_absolute() -> None:
    path = (
        Path(__file__).parent.parent
        / "fixtures"
        / "project_with_multi_constraints_dependency"
    )
    expected = f"demo @ {path.as_uri()}"

    prefix = "/" if sys.platform == "win32" else ""
    requirement = f"demo @ file://{prefix}{path.as_posix()}"
    _test_directory_dependency_pep_508("demo", path, requirement, expected)

    requirement = f"demo @ {path}"
    _test_directory_dependency_pep_508("demo", path, requirement, expected)


def test_directory_dependency_pep_508_localhost() -> None:
    path = (
        Path(__file__).parent.parent
        / "fixtures"
        / "project_with_multi_constraints_dependency"
    )
    prefix = "/" if sys.platform == "win32" else ""
    requirement = f"demo @ file://localhost{prefix}{path.as_posix()}"
    expected = f"demo @ {path.as_uri()}"
    _test_directory_dependency_pep_508("demo", path, requirement, expected)


def test_directory_dependency_pep_508_local_relative() -> None:
    path = Path("..") / "fixtures" / "project_with_multi_constraints_dependency"

    requirement = f"demo @ file://{path.as_posix()}"
    with pytest.raises(ValueError):
        _test_directory_dependency_pep_508("demo", path, requirement)

    requirement = f"demo @ {path}"
    base = Path(__file__).parent
    expected = f"demo @ {(base / path).resolve().as_uri()}"
    _test_directory_dependency_pep_508("demo", path, requirement, expected)


def test_directory_dependency_pep_508_with_subdirectory() -> None:
    path = (
        Path(__file__).parent.parent
        / "fixtures"
        / "project_with_multi_constraints_dependency"
    )
    expected = f"demo @ {path.as_uri()}"

    prefix = "/" if sys.platform == "win32" else ""
    requirement = (
        f"demo @ file://{prefix}{path.parent.as_posix()}#subdirectory={path.name}"
    )
    _test_directory_dependency_pep_508("demo", path, requirement, expected)


def test_directory_dependency_pep_508_extras() -> None:
    path = (
        Path(__file__).parent.parent
        / "fixtures"
        / "project_with_multi_constraints_dependency"
    )
    prefix = "/" if sys.platform == "win32" else ""
    requirement = f"demo[foo,bar] @ file://{prefix}{path.as_posix()}"
    expected = f"demo[bar,foo] @ {path.as_uri()}"
    _test_directory_dependency_pep_508("demo", path, requirement, expected)


def test_directory_dependency_pep_508_with_marker() -> None:
    path = (
        Path(__file__).parent.parent
        / "fixtures"
        / "project_with_multi_constraints_dependency"
    )
    prefix = "/" if sys.platform == "win32" else ""
    requirement = f'demo @ file://{prefix}{path.as_posix()} ; sys_platform == "linux"'
    expected = f'demo @ {path.as_uri()} ; sys_platform == "linux"'
    _test_directory_dependency_pep_508("demo", path, requirement, expected)


@pytest.mark.parametrize(
    "name,path,extras,constraint,expected",
    [
        (
            "my-package",
            SAMPLE_PROJECT,
            None,
            None,
            f"my-package (*) @ {SAMPLE_PROJECT.as_uri()}",
        ),
        (
            "my-package",
            SAMPLE_PROJECT,
            ["db"],
            "1.2",
            f"my-package[db] (1.2) @ {SAMPLE_PROJECT.as_uri()}",
        ),
    ],
)
def test_directory_dependency_string_representation(
    name: str,
    path: Path,
    extras: list[str] | None,
    constraint: str | None,
    expected: str,
) -> None:
    dependency = DirectoryDependency(name=name, path=path, extras=extras)
    if constraint:
        dependency.constraint = constraint
    assert str(dependency) == expected


@pytest.mark.parametrize(
    ("fixture", "name"),
    [
        ("project_with_pep517_non_poetry", "PEP 517"),
        ("project_with_setup_cfg_only", "setup.cfg"),
    ],
)
def test_directory_dependency_non_poetry_pep517(fixture: str, name: str) -> None:
    path = Path(__file__).parent.parent / "fixtures" / fixture

    try:
        DirectoryDependency("package", path)
    except ValueError as e:
        if "does not seem to be a Python package" not in str(e):
            raise e from e
        pytest.fail(f"A {name} project not recognized as valid directory dependency")



from __future__ import annotations

from pathlib import Path

import pytest

from poetry.core.utils.helpers import combine_unicode
from poetry.core.utils.helpers import parse_requires
from poetry.core.utils.helpers import readme_content_type


def test_parse_requires() -> None:
    requires = """\
jsonschema>=2.6.0.0,<3.0.0.0
lockfile>=0.12.0.0,<0.13.0.0
pip-tools>=1.11.0.0,<2.0.0.0
pkginfo>=1.4.0.0,<2.0.0.0
pyrsistent>=0.14.2.0,<0.15.0.0
toml>=0.9.0.0,<0.10.0.0
cleo>=0.6.0.0,<0.7.0.0
cachy>=0.1.1.0,<0.2.0.0
cachecontrol>=0.12.4.0,<0.13.0.0
requests>=2.18.0.0,<3.0.0.0
msgpack-python>=0.5.0.0,<0.6.0.0
pyparsing>=2.2.0.0,<3.0.0.0
requests-toolbelt>=0.8.0.0,<0.9.0.0

[:(python_version >= "2.7.0.0" and python_version < "2.8.0.0") or (python_version >= "3.4.0.0" and python_version < "3.5.0.0")]
typing>=3.6.0.0,<4.0.0.0

[:python_version >= "2.7.0.0" and python_version < "2.8.0.0"]
virtualenv>=15.2.0.0,<16.0.0.0
pathlib2>=2.3.0.0,<3.0.0.0

[:python_version >= "3.4.0.0" and python_version < "3.6.0.0"]
zipfile36>=0.1.0.0,<0.2.0.0

[dev]
isort@ git+git://github.com/timothycrosley/isort.git@e63ae06ec7d70b06df9e528357650281a3d3ec22#egg=isort
"""
    result = parse_requires(requires)
    expected = [
        "jsonschema>=2.6.0.0,<3.0.0.0",
        "lockfile>=0.12.0.0,<0.13.0.0",
        "pip-tools>=1.11.0.0,<2.0.0.0",
        "pkginfo>=1.4.0.0,<2.0.0.0",
        "pyrsistent>=0.14.2.0,<0.15.0.0",
        "toml>=0.9.0.0,<0.10.0.0",
        "cleo>=0.6.0.0,<0.7.0.0",
        "cachy>=0.1.1.0,<0.2.0.0",
        "cachecontrol>=0.12.4.0,<0.13.0.0",
        "requests>=2.18.0.0,<3.0.0.0",
        "msgpack-python>=0.5.0.0,<0.6.0.0",
        "pyparsing>=2.2.0.0,<3.0.0.0",
        "requests-toolbelt>=0.8.0.0,<0.9.0.0",
        (
            'typing>=3.6.0.0,<4.0.0.0 ; (python_version >= "2.7.0.0" and python_version'
            ' < "2.8.0.0") or (python_version >= "3.4.0.0" and python_version <'
            ' "3.5.0.0")'
        ),
        (
            'virtualenv>=15.2.0.0,<16.0.0.0 ; python_version >= "2.7.0.0" and'
            ' python_version < "2.8.0.0"'
        ),
        (
            'pathlib2>=2.3.0.0,<3.0.0.0 ; python_version >= "2.7.0.0" and'
            ' python_version < "2.8.0.0"'
        ),
        (
            'zipfile36>=0.1.0.0,<0.2.0.0 ; python_version >= "3.4.0.0" and'
            ' python_version < "3.6.0.0"'
        ),
        (
            "isort@"
            " git+git://github.com/timothycrosley/isort.git@e63ae06ec7d70b06df9e528357650281a3d3ec22#egg=isort"
            ' ; extra == "dev"'
        ),
    ]
    assert result == expected


def test_utils_helpers_combine_unicode() -> None:
    combined_expected = "é"
    decomposed = "é"
    assert combined_expected != decomposed

    combined = combine_unicode(decomposed)
    assert combined == combined_expected


@pytest.mark.parametrize(
    "readme, content_type",
    [
        ("README.rst", "text/x-rst"),
        ("README.md", "text/markdown"),
        ("README", "text/plain"),
        (Path("README.rst"), "text/x-rst"),
        (Path("README.md"), "text/markdown"),
        (Path("README"), "text/plain"),
    ],
)
def test_utils_helpers_readme_content_type(
    readme: str | Path, content_type: str
) -> None:
    assert readme_content_type(readme) == content_type



from __future__ import annotations

import pytest

from packaging.utils import canonicalize_name

from poetry.core.constraints.version.exceptions import ParseConstraintError
from poetry.core.packages.dependency import Dependency
from poetry.core.packages.directory_dependency import DirectoryDependency
from poetry.core.packages.file_dependency import FileDependency
from poetry.core.version.markers import InvalidMarkerError
from poetry.core.version.markers import parse_marker
from poetry.core.version.requirements import InvalidRequirementError


@pytest.mark.parametrize(
    "constraint",
    [
        "^1.0",
        "^1.0.dev0",
        "^1.0.0",
        "^1.0.0.dev0",
        "^1.0.0.alpha0",
        "^1.0.0.alpha0+local",
        "^1.0.0.rc0+local",
        "^1.0.0-1",
    ],
)
@pytest.mark.parametrize("allows_prereleases", [None, False, True])
def test_allows_prerelease(constraint: str, allows_prereleases: bool) -> None:
    dependency = Dependency("A", constraint, allows_prereleases=allows_prereleases)
    assert dependency.allows_prereleases() == allows_prereleases


def test_python_versions_are_made_precise() -> None:
    dependency = Dependency("Django", "^1.23")
    dependency.python_versions = ">3.6,<=3.10"

    assert (
        str(dependency.marker)
        == 'python_full_version > "3.6.0" and python_full_version <= "3.10.0"'
    )
    assert str(dependency.python_constraint) == ">3.6,<=3.10"


def test_to_pep_508() -> None:
    dependency = Dependency("Django", "^1.23")

    result = dependency.to_pep_508()
    assert result == "Django (>=1.23,<2.0)"


def test_to_pep_508_wilcard() -> None:
    dependency = Dependency("Django", "*")

    result = dependency.to_pep_508()
    assert result == "Django"


def test_to_pep_508_in_extras() -> None:
    dependency = Dependency("Django", "^1.23")
    dependency._in_extras = [canonicalize_name("foo")]

    result = dependency.to_pep_508()
    assert result == 'Django (>=1.23,<2.0) ; extra == "foo"'

    result = dependency.to_pep_508(with_extras=False)
    assert result == "Django (>=1.23,<2.0)"

    dependency._in_extras = [canonicalize_name("foo"), canonicalize_name("bar")]

    result = dependency.to_pep_508()
    assert result == 'Django (>=1.23,<2.0) ; extra == "foo" or extra == "bar"'

    dependency.python_versions = "~2.7 || ^3.6"

    result = dependency.to_pep_508()
    assert (
        result == "Django (>=1.23,<2.0) ; "
        "("
        'python_version == "2.7" '
        'or python_version >= "3.6" and python_version < "4.0"'
        ") "
        'and (extra == "foo" or extra == "bar")'
    )

    result = dependency.to_pep_508(with_extras=False)
    assert (
        result == "Django (>=1.23,<2.0) ; "
        'python_version == "2.7" '
        'or python_version >= "3.6" and python_version < "4.0"'
    )


def test_to_pep_508_in_extras_parsed() -> None:
    dependency = Dependency.create_from_pep_508(
        'foo[baz,bar] (>=1.23,<2.0) ; extra == "baz"'
    )

    result = dependency.to_pep_508()
    assert result == 'foo[bar,baz] (>=1.23,<2.0) ; extra == "baz"'

    result = dependency.to_pep_508(with_extras=False)
    assert result == "foo[bar,baz] (>=1.23,<2.0)"


@pytest.mark.parametrize(
    ("exclusion", "expected"),
    [
        ("!=1.2.3", "!=1.2.3"),
        ("!=1.2.*", "!=1.2.*"),
        ("<2.0 || >=2.1.dev0", "!=2.0.*"),
    ],
)
def test_to_pep_508_with_excluded_versions(exclusion: str, expected: str) -> None:
    dependency = Dependency("foo", exclusion)

    assert dependency.to_pep_508() == f"foo ({expected})"


@pytest.mark.parametrize(
    "python_versions, marker",
    [
        (">=3.5,<3.5.4", 'python_version >= "3.5" and python_full_version < "3.5.4"'),
        (">=3.5.4,<3.6", 'python_full_version >= "3.5.4" and python_version < "3.6"'),
        ("<3.5.4", 'python_full_version < "3.5.4"'),
        (">=3.5.4", 'python_full_version >= "3.5.4"'),
        ("== 3.5.4", 'python_full_version == "3.5.4"'),
        (">=3,<4", 'python_version >= "3" and python_version < "4"'),
        ("==3.*", 'python_version >= "3" and python_version < "4"'),
        (
            "~2.7 || ^3.6",
            'python_version == "2.7" or python_version >= "3.6" and python_version < "4.0"',
        ),
    ],
)
def test_to_pep_508_with_python_versions(python_versions: str, marker: str) -> None:
    dependency = Dependency("Django", "^1.23")
    dependency.python_versions = python_versions

    expected = f"Django (>=1.23,<2.0) ; {marker}"

    assert dependency.to_pep_508() == expected
    assert dependency.to_pep_508(resolved=True) == expected
    assert str(dependency.marker) == marker


def test_to_pep_508_tilde() -> None:
    dependency = Dependency("foo", "~1.2.3")

    assert dependency.to_pep_508() == "foo (>=1.2.3,<1.3.0)"

    dependency = Dependency("foo", "~1.2")

    assert dependency.to_pep_508() == "foo (>=1.2,<1.3)"

    dependency = Dependency("foo", "~0.2.3")

    assert dependency.to_pep_508() == "foo (>=0.2.3,<0.3.0)"

    dependency = Dependency("foo", "~0.2")

    assert dependency.to_pep_508() == "foo (>=0.2,<0.3)"


def test_to_pep_508_caret() -> None:
    dependency = Dependency("foo", "^1.2.3")

    assert dependency.to_pep_508() == "foo (>=1.2.3,<2.0.0)"

    dependency = Dependency("foo", "^1.2")

    assert dependency.to_pep_508() == "foo (>=1.2,<2.0)"

    dependency = Dependency("foo", "^0.2.3")

    assert dependency.to_pep_508() == "foo (>=0.2.3,<0.3.0)"

    dependency = Dependency("foo", "^0.2")

    assert dependency.to_pep_508() == "foo (>=0.2,<0.3)"


def test_to_pep_508_combination() -> None:
    dependency = Dependency("foo", "^1.2,!=1.3.5")

    assert dependency.to_pep_508() == "foo (>=1.2,<2.0,!=1.3.5)"

    dependency = Dependency("foo", "~1.2,!=1.2.5")

    assert dependency.to_pep_508() == "foo (>=1.2,<1.3,!=1.2.5)"


@pytest.mark.parametrize(
    "requirement",
    [
        "enum34; extra == ':python_version < \"3.4\"'",
        "enum34; extra == \":python_version < '3.4'\"",
    ],
)
def test_to_pep_508_with_invalid_marker(requirement: str) -> None:
    with pytest.raises(InvalidMarkerError):
        _ = Dependency.create_from_pep_508(requirement)


@pytest.mark.parametrize(
    "requirement",
    [
        'enum34; extra == ":python_version < "3.4""',
    ],
)
def test_to_pep_508_with_invalid_requirement(requirement: str) -> None:
    with pytest.raises(InvalidRequirementError):
        _ = Dependency.create_from_pep_508(requirement)


@pytest.mark.parametrize(
    ("requirement", "dependency_type"),
    [
        (
            "eflips-depot @ git@github.com/mpm-tu-berlin/eflips-depot.git@feature/allow-only-oppo-charging",
            DirectoryDependency,
        ),
        (
            "eflips-depot @ git@github.com/mpm-tu-berlin/eflips-depot.git@feature/allow-only-oppo-charging.whl",
            FileDependency,
        ),
    ],
)
def test_to_pep_508_with_invalid_path_requirement(
    requirement: str, dependency_type: type[FileDependency | DirectoryDependency]
) -> None:
    dependency = Dependency.create_from_pep_508(requirement)
    assert isinstance(dependency, dependency_type)
    assert dependency.source_url


def test_complete_name() -> None:
    assert Dependency("foo", ">=1.2.3").complete_name == "foo"
    assert (
        Dependency("foo", ">=1.2.3", extras=["baz", "bar"]).complete_name
        == "foo[bar,baz]"
    )


@pytest.mark.parametrize(
    "name,constraint,extras,expected",
    [
        ("A", ">2.7,<3.0", None, "A (>2.7,<3.0)"),
        ("A", ">2.7,<3.0", ["x"], "A[x] (>2.7,<3.0)"),
        ("A", ">=1.6.5,<1.8.0 || >1.8.0,<3.1.0", None, "A (>=1.6.5,!=1.8.0,<3.1.0)"),
        (
            "A",
            ">=1.6.5,<1.8.0 || >1.8.0,<3.1.0",
            ["x"],
            "A[x] (>=1.6.5,!=1.8.0,<3.1.0)",
        ),
        # test single version range (wildcard)
        ("A", "==2.*", None, "A (==2.*)"),
        ("A", "==2.0.*", None, "A (==2.0.*)"),
        ("A", "==0.0.*", None, "A (==0.0.*)"),
        ("A", "==0.1.*", None, "A (==0.1.*)"),
        ("A", "==0.*", None, "A (==0.*)"),
        ("A", ">=1.0.dev0,<2", None, "A (==1.*)"),
        ("A", ">=1.dev0,<2", None, "A (==1.*)"),
        ("A", ">=1.0.dev1,<2", None, "A (>=1.0.dev1,<2)"),
        ("A", ">=1.1.dev0,<2", None, "A (>=1.1.dev0,<2)"),
        ("A", ">=1.0.dev0,<2.0.dev0", None, "A (==1.*)"),
        ("A", ">=1.0.dev0,<2.0.dev1", None, "A (>=1.0.dev0,<2.0.dev1)"),
        ("A", ">=1,<2", None, "A (>=1,<2)"),
        ("A", ">=1.0.dev0,<1.1", None, "A (==1.0.*)"),
        ("A", ">=1.0.0.0.dev0,<1.1.0.0.0", None, "A (==1.0.*)"),
        # test single version range (wildcard) exclusions
        ("A", ">=1.8,!=2.0.*", None, "A (>=1.8,!=2.0.*)"),
        ("A", "!=0.0.*", None, "A (!=0.0.*)"),
        ("A", "!=0.1.*", None, "A (!=0.1.*)"),
        ("A", "!=0.*", None, "A (!=0.*)"),
        ("A", ">=1.8,!=2.*", None, "A (>=1.8,!=2.*)"),
        ("A", ">=1.8,!=2.*.*", None, "A (>=1.8,!=2.*)"),
        ("A", ">=1.8,<2.0 || >=2.1.0.dev0", None, "A (>=1.8,!=2.0.*)"),
        ("A", ">=1.8,<2.0.0 || >=3.0.0.dev0", None, "A (>=1.8,!=2.*)"),
        ("A", ">=1.8,<2.0 || >=3.dev0", None, "A (>=1.8,!=2.*)"),
        ("A", ">=1.8,<2 || >=2.1.0.dev0", None, "A (>=1.8,!=2.0.*)"),
        ("A", ">=1.8,<2 || >=2.1.dev0", None, "A (>=1.8,!=2.0.*)"),
        ("A", ">=1.8,!=2.0.*,!=3.0.*", None, "A (>=1.8,!=2.0.*,!=3.0.*)"),
        ("A", ">=1.8.0.0,<2.0.0.0 || >=2.0.1.0.dev0", None, "A (>=1.8.0.0,!=2.0.0.*)"),
        ("A", ">=1.8.0.0,<2 || >=2.0.1.0.dev0", None, "A (>=1.8.0.0,!=2.0.0.*)"),
        # we verify that the range exclusion logic is not too eager
        ("A", ">=1.8,<2.0 || >=2.2.0", None, "A (>=1.8,<2.0 || >=2.2.0)"),
        ("A", ">=1.8,<2.0 || >=2.1.5", None, "A (>=1.8,<2.0 || >=2.1.5)"),
        ("A", ">=1.8.0.0,<2 || >=2.0.1.5", None, "A (>=1.8.0.0,<2 || >=2.0.1.5)"),
        ("A", ">=1.8.0.0,!=2.0.0.*", None, "A (>=1.8.0.0,!=2.0.0.*)"),
    ],
)
def test_dependency_string_representation(
    name: str, constraint: str, extras: list[str] | None, expected: str
) -> None:
    dependency = Dependency(name=name, constraint=constraint, extras=extras)
    assert str(dependency) == expected


def test_set_constraint_sets_pretty_constraint() -> None:
    dependency = Dependency("A", "^1.0")
    assert dependency.pretty_constraint == "^1.0"
    dependency.constraint = "^2.0"
    assert dependency.pretty_constraint == "^2.0"


def test_set_bogus_constraint_raises_exception() -> None:
    dependency = Dependency("A", "^1.0")
    with pytest.raises(ParseConstraintError):
        dependency.constraint = "^=4.5"


def test_with_constraint() -> None:
    dependency = Dependency(
        "foo",
        "^1.2.3",
        optional=True,
        groups=["DEV"],
        allows_prereleases=True,
        extras=["bar", "baz"],
    )
    dependency.marker = parse_marker(
        'python_version >= "3.6" and python_version < "4.0"'
    )
    dependency.transitive_marker = parse_marker(
        'python_version >= "3.7" and python_version < "4.0"'
    )
    dependency.python_versions = "^3.6"

    new = dependency.with_constraint("^1.2.6")

    assert new.name == dependency.name
    assert str(new.constraint) == ">=1.2.6,<2.0.0"
    assert str(dependency.constraint) == ">=1.2.3,<2.0.0"
    assert new.is_optional()
    assert new.groups == frozenset(["dev"])
    assert new.allows_prereleases()
    assert set(new.extras) == {"bar", "baz"}
    assert new.marker == dependency.marker
    assert new.transitive_marker == dependency.transitive_marker
    assert new.python_constraint == dependency.python_constraint


def test_with_groups() -> None:
    dependency = Dependency("foo", "^1.2.3", groups=["DEV"])

    new = dependency.with_groups(["DOC", "test"])

    assert new.name == dependency.name
    assert str(dependency.constraint) == ">=1.2.3,<2.0.0"
    assert new.groups == frozenset(["doc", "test"])
    assert dependency.groups == frozenset(["dev"])


@pytest.mark.parametrize(
    "marker, expected",
    [
        ('python_version >= "3.6" and python_version < "4.0"', ">=3.6,<4.0"),
        ('sys_platform == "linux"', "*"),
        ('python_version >= "3.9" or sys_platform == "linux"', "*"),
        ('python_version >= "3.9" and sys_platform == "linux"', ">=3.9"),
    ],
)
def test_marker_properly_sets_python_constraint(marker: str, expected: str) -> None:
    dependency = Dependency("foo", "^1.2.3")
    dependency.marker = marker
    assert str(dependency.python_constraint) == expected


def test_dependency_markers_are_the_same_as_markers() -> None:
    dependency = Dependency.create_from_pep_508('foo ; extra=="bar"')
    marker = parse_marker('extra=="bar"')

    assert dependency.marker == marker


def test_marker_properly_unsets_python_constraint() -> None:
    dependency = Dependency("foo", "^1.2.3")

    dependency.marker = 'python_version >= "3.6"'
    assert str(dependency.python_constraint) == ">=3.6"

    dependency.marker = "*"
    assert str(dependency.python_constraint) == "*"


def test_create_from_pep_508_url_with_activated_extras() -> None:
    dependency = Dependency.create_from_pep_508("name [fred,bar] @ http://foo.com")
    assert dependency.extras == {"fred", "bar"}


def test_create_from_pep_508_starting_with_digit() -> None:
    dependency = Dependency.create_from_pep_508("2captcha-python")
    assert dependency.name == "2captcha-python"


@pytest.mark.parametrize(
    "dependency1, dependency2, expected",
    [
        (Dependency("a", "1.0"), Dependency("a", "1.0"), True),
        (Dependency("a", "1.0"), Dependency("a", "1.0.1"), False),
        (Dependency("a", "1.0"), Dependency("a1", "1.0"), False),
        (Dependency("a", "1.0"), Dependency("a", "1.0", source_type="file"), False),
        # constraint is implicitly given for direct origin dependencies,
        # but might not be set
        (
            Dependency("a", "1.0", source_type="file"),
            Dependency("a", "*", source_type="file"),
            True,
        ),
        # constraint is not implicit for non direct origin dependencies
        (Dependency("a", "1.0"), Dependency("a", "*"), False),
        (
            Dependency("a", "1.0", source_type="legacy"),
            Dependency("a", "*", source_type="legacy"),
            False,
        ),
    ],
)
def test_eq(dependency1: Dependency, dependency2: Dependency, expected: bool) -> None:
    assert (dependency1 == dependency2) is expected
    assert (dependency2 == dependency1) is expected


@pytest.mark.parametrize(
    "attr_name, value",
    [
        ("constraint", "2.0"),
        ("python_versions", "<3.8"),
        ("marker", "sys_platform == 'linux'"),
        ("transitive_marker", "sys_platform == 'linux'"),
    ],
)
def test_mutable_attributes_not_in_hash(attr_name: str, value: str) -> None:
    dependency = Dependency("foo", "^1.2.3")
    ref_hash = hash(dependency)
    ref_value = getattr(dependency, attr_name)

    setattr(dependency, attr_name, value)

    assert value != ref_value
    assert hash(dependency) == ref_hash



from __future__ import annotations

from poetry.core.constraints.generic.any_constraint import AnyConstraint
from poetry.core.constraints.generic.base_constraint import BaseConstraint
from poetry.core.constraints.generic.constraint import Constraint
from poetry.core.constraints.generic.empty_constraint import EmptyConstraint
from poetry.core.constraints.generic.multi_constraint import MultiConstraint
from poetry.core.constraints.generic.parser import parse_constraint
from poetry.core.constraints.generic.parser import parse_extra_constraint
from poetry.core.constraints.generic.union_constraint import UnionConstraint


__all__ = (
    "AnyConstraint",
    "BaseConstraint",
    "Constraint",
    "EmptyConstraint",
    "MultiConstraint",
    "UnionConstraint",
    "parse_constraint",
    "parse_extra_constraint",
)



from __future__ import annotations

from poetry.core.constraints.version.empty_constraint import EmptyConstraint
from poetry.core.constraints.version.parser import parse_constraint
from poetry.core.constraints.version.parser import parse_marker_version_constraint
from poetry.core.constraints.version.util import constraint_regions
from poetry.core.constraints.version.version import Version
from poetry.core.constraints.version.version_constraint import VersionConstraint
from poetry.core.constraints.version.version_range import VersionRange
from poetry.core.constraints.version.version_range_constraint import (
    VersionRangeConstraint,
)
from poetry.core.constraints.version.version_union import VersionUnion


__all__ = (
    "EmptyConstraint",
    "Version",
    "VersionConstraint",
    "VersionRange",
    "VersionRangeConstraint",
    "VersionUnion",
    "constraint_regions",
    "parse_constraint",
    "parse_marker_version_constraint",
)



from __future__ import annotations

import shutil

from pathlib import Path

import pytest


fixtures_dir = Path(__file__).parent / "fixtures"


@pytest.fixture
def complete_with_pycache_and_pyc_files(tmp_path: Path) -> Path:
    root = fixtures_dir / "complete"
    tmp_root = tmp_path / "complete"  # not git repo!

    shutil.copytree(root, tmp_root)
    for location in (".", "sub_pkg1"):
        abs_location = tmp_root / "my_package" / location
        (abs_location / "module1.cpython-39.pyc").touch()
        pycache_location = tmp_root / "my_package" / location / "__pycache__"
        pycache_location.mkdir(parents=True)
        (pycache_location / "module1.cpython-39.pyc").touch()
        (pycache_location / "some_other_file").touch()

    return tmp_root
